"""Geocoding module for getting a geometry from query using Nominatim."""

import hashlib
import json
from pathlib import Path
from typing import Any, Optional, Union, cast, overload

from geopy.geocoders.nominatim import Nominatim
from geopy.location import Location
from shapely.geometry import shape
from shapely.geometry.base import BaseGeometry
from shapely.ops import unary_union

from overturemaestro._exceptions import QueryNotGeocodedError


@overload
def geocode_to_geometry(query: str) -> BaseGeometry: ...


@overload
def geocode_to_geometry(query: list[str]) -> BaseGeometry: ...


def geocode_to_geometry(query: Union[str, list[str]]) -> BaseGeometry:
    """Geocode a query to a (Multi)Polygon geometry using Nominatim."""
    if not isinstance(query, str):
        return unary_union([geocode_to_geometry(sub_query) for sub_query in query])

    h = hashlib.new("sha256")
    h.update(query.encode())
    query_hash = h.hexdigest()
    query_file_path = Path("cache").resolve() / f"{query_hash}.json"

    if not query_file_path.exists():
        query_results = Nominatim(
            user_agent="OvertureMaestro Python package (https://github.com/kraina-ai/overturemaestro)"
        ).geocode(query, geometry="geojson", exactly_one=False)

        if not query_results:
            raise QueryNotGeocodedError(f"Zero results from Nominatim for query '{query}'.")

        polygon_result = _get_first_polygon(query_results)

        if not polygon_result:
            raise QueryNotGeocodedError(f"No polygon found for query '{query}'.")

        query_file_path.parent.mkdir(parents=True, exist_ok=True)
        query_file_path.write_text(json.dumps(polygon_result))
    else:
        polygon_result = json.loads(query_file_path.read_text())

    return unary_union(shape(polygon_result))


def _get_first_polygon(results: list[Location]) -> Optional[dict[str, Any]]:
    """
    Choose first result of geometry type (Multi)Polygon from list of results.

    Inspired by OSMnx implementation.
    """
    polygon_types = {"Polygon", "MultiPolygon"}

    for result in results:
        geojson_dict = cast(dict[str, Any], result.raw["geojson"])
        if geojson_dict["type"] in polygon_types:
            return geojson_dict

    return None



"""Helper functions for DuckDB."""

from pathlib import Path
from typing import Union

import duckdb


def _sql_escape(value: str) -> str:
    """Escape value for SQL query."""
    return value.replace("'", "''")


def _set_up_duckdb_connection(tmp_dir_path: Union[str, Path]) -> "duckdb.DuckDBPyConnection":
    """Create DuckDB connection in a given directory."""
    local_db_file = "db.duckdb"
    connection = duckdb.connect(
        database=str(Path(tmp_dir_path) / local_db_file),
        config=dict(preserve_insertion_order=False),
    )
    connection.sql("SET enable_progress_bar = false;")
    connection.sql("SET enable_progress_bar_print = false;")

    connection.install_extension("spatial")
    connection.load_extension("spatial")

    return connection



"""CLI module for the OvertureMaestro functions."""

import logging
from pathlib import Path
from typing import TYPE_CHECKING, Annotated, Optional, cast

import click
import typer

from overturemaestro._geopandas_api_version import GEOPANDAS_NEW_API

if TYPE_CHECKING:  # pragma: no cover
    from overturemaestro._rich_progress import VERBOSITY_MODE

app = typer.Typer(context_settings={"help_option_names": ["-h", "--help"]}, rich_markup_mode="rich")


def _version_callback(value: bool) -> None:
    if value:
        from overturemaestro import __app_name__, __version__

        typer.echo(f"{__app_name__} {__version__}")
        raise typer.Exit()


def _display_release_versions_callback(ctx: typer.Context, value: bool) -> None:
    if value:
        from rich import print as rprint
        from rich.table import Table

        from overturemaestro import get_available_release_versions

        release_versions = get_available_release_versions()

        table = Table()
        table.add_column("Release Version", no_wrap=True)
        for release_version in release_versions:
            table.add_row(release_version)

        rprint(table)

        raise typer.Exit()


def _display_theme_type_pairs_callback(ctx: typer.Context, value: bool) -> None:
    if value:
        from rich import print as rprint
        from rich.table import Table

        from overturemaestro import get_available_theme_type_pairs, get_newest_release_version

        param_values = {p.name: p.default for p in ctx.command.params}
        param_values.update(ctx.params)
        release_version = cast(
            str, param_values.get("release_version") or get_newest_release_version()
        )
        theme_type_pairs = get_available_theme_type_pairs(release=release_version)

        table = Table(title=f"{release_version} release")
        table.add_column("Theme", no_wrap=True)
        table.add_column("Type", no_wrap=True)
        for theme_value, type_value in theme_type_pairs:
            table.add_row(theme_value, type_value)

        rprint(table)
        raise typer.Exit()


def _path_callback(ctx: typer.Context, value: Path) -> Path:
    if not Path(value).exists():
        raise typer.BadParameter(f"File not found error: {value}")
    return value


def _empty_path_callback(ctx: typer.Context, value: Path) -> Optional[Path]:
    if not value:
        return None
    return _path_callback(ctx, value)


class BboxGeometryParser(click.ParamType):  # type: ignore
    """Parser for geometry in bounding box form."""

    name = "BBOX"

    def convert(self, value, param=None, ctx=None):  # type: ignore
        """Convert parameter value."""
        try:
            from shapely import box

            bbox_values = [float(x.strip()) for x in value.split(",")]
            return box(*bbox_values)
        except ValueError:  # ValueError raised when passing non-numbers to float()
            raise typer.BadParameter(
                "Cannot parse provided bounding box."
                " Valid value must contain 4 floating point numbers"
                " separated by commas."
            ) from None


class WktGeometryParser(click.ParamType):  # type: ignore
    """Parser for geometry in WKT form."""

    name = "TEXT (WKT)"

    def convert(self, value, param=None, ctx=None):  # type: ignore
        """Convert parameter value."""
        if not value:
            return None
        try:
            from shapely import from_wkt

            return from_wkt(value)
        except Exception:
            raise typer.BadParameter("Cannot parse provided WKT") from None


class GeoJsonGeometryParser(click.ParamType):  # type: ignore
    """Parser for geometry in GeoJSON form."""

    name = "TEXT (GeoJSON)"

    def convert(self, value, param=None, ctx=None):  # type: ignore
        """Convert parameter value."""
        if not value:
            return None
        try:
            from shapely import from_geojson

            return from_geojson(value)
        except Exception:
            raise typer.BadParameter("Cannot parse provided GeoJSON") from None


class GeoFileGeometryParser(click.ParamType):  # type: ignore
    """Parser for geometry in geo file form."""

    name = "PATH"

    def convert(self, value, param=None, ctx=None):  # type: ignore
        """Convert parameter value."""
        if not value:
            return None

        value = _path_callback(ctx=ctx, value=value)

        try:
            import geopandas as gpd

            gdf = gpd.read_file(value)
            if GEOPANDAS_NEW_API:
                return gdf.union_all()
            else:
                return gdf.unary_union
        except Exception:
            raise typer.BadParameter("Cannot parse provided geo file") from None


class GeocodeGeometryParser(click.ParamType):  # type: ignore
    """Parser for geometry in string Nominatim query form."""

    name = "TEXT"

    def convert(self, value, param=None, ctx=None):  # type: ignore
        """Convert parameter value."""
        if not value:
            return None

        try:
            from overturemaestro.geocode import geocode_to_geometry

            return geocode_to_geometry(value)
        except Exception:
            raise typer.BadParameter("Cannot geocode provided Nominatim query") from None


class GeohashGeometryParser(click.ParamType):  # type: ignore
    """Parser for geometry in string Nominatim query form."""

    name = "TEXT (Geohash)"

    def convert(self, value, param=None, ctx=None):  # type: ignore
        """Convert parameter value."""
        if not value:
            return None

        try:
            import geopandas as gpd
            from geohash import bbox as geohash_bbox
            from shapely.geometry import box

            geometries = []
            for geohash in value.split(","):
                bounds = geohash_bbox(geohash.strip())
                geometries.append(
                    box(minx=bounds["w"], miny=bounds["s"], maxx=bounds["e"], maxy=bounds["n"])
                )
            if GEOPANDAS_NEW_API:
                return gpd.GeoSeries(geometries).union_all()
            else:
                return gpd.GeoSeries(geometries).unary_union
        except Exception:
            raise
            # raise typer.BadParameter(f"Cannot parse provided Geohash value: {geohash}") from None


class H3GeometryParser(click.ParamType):  # type: ignore
    """Parser for geometry in string Nominatim query form."""

    name = "TEXT (H3)"

    def convert(self, value, param=None, ctx=None):  # type: ignore
        """Convert parameter value."""
        if not value:
            return None

        try:
            import geopandas as gpd
            import h3
            from shapely.geometry import Polygon

            geometries = []  # noqa: FURB138
            for h3_cell in value.split(","):
                geometries.append(
                    Polygon([coords[::-1] for coords in h3.cell_to_boundary(h3_cell.strip())])
                )
            if GEOPANDAS_NEW_API:
                return gpd.GeoSeries(geometries).union_all()
            else:
                return gpd.GeoSeries(geometries).unary_union
        except Exception as ex:
            raise typer.BadParameter(f"Cannot parse provided H3 values: {value}") from ex


class S2GeometryParser(click.ParamType):  # type: ignore
    """Parser for geometry in string Nominatim query form."""

    name = "TEXT (S2)"

    def convert(self, value, param=None, ctx=None):  # type: ignore
        """Convert parameter value."""
        if not value:
            return None

        try:
            import geopandas as gpd
            from s2 import s2
            from shapely.geometry import Polygon

            geometries = []  # noqa: FURB138
            for s2_index in value.split(","):
                geometries.append(
                    Polygon(s2.s2_to_geo_boundary(s2_index.strip(), geo_json_conformant=True))
                )
            if GEOPANDAS_NEW_API:
                return gpd.GeoSeries(geometries).union_all()
            else:
                return gpd.GeoSeries(geometries).unary_union
        except Exception:
            raise typer.BadParameter(f"Cannot parse provided S2 value: {s2_index}") from None


class PyArrowExpressionParser(click.ParamType):  # type: ignore
    """Parser for geometry in string Nominatim query form."""

    name = "FILTER"

    def convert(self, value, param=None, ctx=None):  # type: ignore
        """Convert parameter value."""
        if not value:
            return None

        try:
            parts = value.split()

            if len(parts) != 3:
                raise typer.BadParameter(
                    "Provided expression is not in a required format (3 elements in a string)."
                ) from None

            from pyarrow.parquet import filters_to_expression

            # Parse numbers
            if parts[2].replace(".", "", 1).isdigit():
                parts[2] = float(parts[2])

            # TODO: optional - add support for "in" operator with list of values

            # Check if columns are nested
            column_part = parts[0]
            if "." in column_part:
                columns = tuple(column_part.split("."))
                parts = [[columns, parts[1], parts[2]]]
            elif "," in column_part:
                import warnings

                warnings.warn(
                    (
                        "Found columns split by a comma. New suggested format are"
                        " column names separated by a dot. For compatilibity reasons"
                        " OvertureMaestro will split columns separated by a comma,"
                        " but it will result in this warning."
                    ),
                    DeprecationWarning,
                    stacklevel=0,
                )
                columns = tuple(column_part.split(","))
                parts = [[columns, parts[1], parts[2]]]

            pyarrow_filter = filters_to_expression([parts])
            return pyarrow_filter

        except Exception as ex:
            raise typer.BadParameter(
                f"Cannot parse provided PyArrow Expression: {value}."
                " Required format: <column_name> <operator> <value>."
            ) from ex


@app.command()  # type: ignore
def main(
    theme_value: Annotated[
        Optional[str],
        typer.Argument(
            help="Data [bold yellow]theme[/bold yellow] value",
            metavar="theme",
            show_default=False,
        ),
    ] = None,
    type_value: Annotated[
        Optional[str],
        typer.Argument(
            help="Feature [bold yellow]type[/bold yellow] within [yellow]theme[/yellow]",
            metavar="type",
            show_default=False,
        ),
    ] = None,
    release_version: Annotated[
        Optional[str],
        typer.Option(
            "--release-version",
            "--release",
            help=(
                "OvertureMaps dataset release version."
                " If not provided, will automatically select the newest"
                " available version."
            ),
            is_eager=True,
            show_default=False,
        ),
    ] = None,
    geom_filter_bbox: Annotated[
        Optional[str],
        typer.Option(
            help=(
                "Geometry to use as a filter in the"
                " [bold dark_orange]bounding box[/bold dark_orange] format - 4 floating point"
                " numbers separated by commas."
                " Cannot be used together with other"
                " [bold bright_cyan]geom-filter-...[/bold bright_cyan] parameters."
            ),
            click_type=BboxGeometryParser(),
            show_default=False,
        ),
    ] = None,
    geom_filter_file: Annotated[
        Optional[str],
        typer.Option(
            help=(
                "Geometry to use as a filter in the"
                " [bold dark_orange]file[/bold dark_orange] format - any that can be opened by"
                " GeoPandas. Will return the unary union of the geometries in the file."
                " Cannot be used together with other"
                " [bold bright_cyan]geom-filter-...[/bold bright_cyan] parameters."
            ),
            click_type=GeoFileGeometryParser(),
            show_default=False,
        ),
    ] = None,
    geom_filter_geocode: Annotated[
        Optional[str],
        typer.Option(
            help=(
                "Geometry to use as a filter in the"
                " [bold dark_orange]string to geocode[/bold dark_orange] format - it will be"
                " geocoded to the geometry using Nominatim API (GeoPy library)."
                " Cannot be used together with other"
                " [bold bright_cyan]geom-filter-...[/bold bright_cyan] parameters."
            ),
            click_type=GeocodeGeometryParser(),
            show_default=False,
        ),
    ] = None,
    geom_filter_geojson: Annotated[
        Optional[str],
        typer.Option(
            help=(
                "Geometry to use as a filter in the [bold dark_orange]GeoJSON[/bold dark_orange]"
                " format."
                " Cannot be used together with other"
                " [bold bright_cyan]geom-filter-...[/bold bright_cyan] parameters."
            ),
            click_type=GeoJsonGeometryParser(),
            show_default=False,
        ),
    ] = None,
    geom_filter_index_geohash: Annotated[
        Optional[str],
        typer.Option(
            help=(
                "Geometry to use as a filter in the"
                " [bold dark_orange]Geohash index[/bold dark_orange]"
                " format. Separate multiple values with a comma."
                " Cannot be used together with other"
                " [bold bright_cyan]geom-filter-...[/bold bright_cyan] parameters."
            ),
            click_type=GeohashGeometryParser(),
            show_default=False,
        ),
    ] = None,
    geom_filter_index_h3: Annotated[
        Optional[str],
        typer.Option(
            help=(
                "Geometry to use as a filter in the [bold dark_orange]H3 index[/bold dark_orange]"
                " format. Separate multiple values with a comma."
                " Cannot be used together with other"
                " [bold bright_cyan]geom-filter-...[/bold bright_cyan] parameters."
            ),
            click_type=H3GeometryParser(),
            show_default=False,
        ),
    ] = None,
    geom_filter_index_s2: Annotated[
        Optional[str],
        typer.Option(
            help=(
                "Geometry to use as a filter in the [bold dark_orange]S2 index[/bold dark_orange]"
                " format. Separate multiple values with a comma."
                " Cannot be used together with other"
                " [bold bright_cyan]geom-filter-...[/bold bright_cyan] parameters."
            ),
            click_type=S2GeometryParser(),
            show_default=False,
        ),
    ] = None,
    geom_filter_wkt: Annotated[
        Optional[str],
        typer.Option(
            help=(
                "Geometry to use as a filter in the [bold dark_orange]WKT[/bold dark_orange]"
                " format."
                " Cannot be used together with other"
                " [bold bright_cyan]geom-filter-...[/bold bright_cyan] parameters."
            ),
            click_type=WktGeometryParser(),
            show_default=False,
        ),
    ] = None,
    pyarrow_filters: Annotated[
        Optional[list[str]],
        typer.Option(
            "--filter",
            "--pyarrow-filter",
            help=(
                "Filters to apply on a pyarrow dataset."
                " Required format: <column(s)> <operator> <value>."
                " Nested column names should be passed separated by a dot."
                " Can pass multiple filters."
            ),
            click_type=PyArrowExpressionParser(),
            show_default=False,
        ),
    ] = None,
    result_file_path: Annotated[
        Optional[Path],
        typer.Option(
            "--output",
            "-o",
            help=(
                "Path where to save final geoparquet file. If not provided, it will be generated"
                " automatically based on the input pbf file name."
            ),
            show_default=False,
        ),
    ] = None,
    ignore_cache: Annotated[
        bool,
        typer.Option(
            "--ignore-cache/",
            "--no-cache/",
            help="Whether to ignore previously precalculated geoparquet files or not.",
            show_default=False,
        ),
    ] = False,
    working_directory: Annotated[
        Path,
        typer.Option(
            "--working-directory",
            "--work-dir",
            help=(
                "Directory where to save the parsed parquet and geoparquet files."
                " Will be created if doesn't exist."
            ),
        ),
    ] = "files",  # type: ignore
    silent_mode: Annotated[
        bool,
        typer.Option(
            "--silent/",
            help="Whether to disable progress reporting.",
            show_default=False,
        ),
    ] = False,
    transient_mode: Annotated[
        bool,
        typer.Option(
            "--transient/",
            help="Whether to make more transient (concise) progress reporting.",
            show_default=False,
        ),
    ] = False,
    # allow_uncovered_geometry: Annotated[
    #     bool,
    #     typer.Option(
    #         "--allow-uncovered-geometry/",
    #         help=(
    #             "Suppresses an error if some geometry parts aren't covered by any OSM extract."
    #             " Works only when PbfFileReader is asked to download OSM extracts automatically."
    #         ),
    #         show_default=False,
    #     ),
    # ] = False,
    show_release_versions: Annotated[
        Optional[bool],
        typer.Option(
            "--show-release-versions",
            help="Show available OvertureMaps release versions and exit.",
            callback=_display_release_versions_callback,
            is_eager=False,
        ),
    ] = None,
    show_theme_type_pairs: Annotated[
        Optional[bool],
        typer.Option(
            "--show-theme-type-pairs",
            help="Show available OvertureMaps theme type pairs for the release and exit.",
            callback=_display_theme_type_pairs_callback,
            is_eager=False,
        ),
    ] = None,
    version: Annotated[
        Optional[bool],
        typer.Option(
            "--version",
            "-v",
            help="Show the application's version and exit.",
            callback=_version_callback,
            is_eager=True,
        ),
    ] = None,
) -> None:
    """
    OvertureMaestro CLI.

    Wraps public functions and prints final path to the saved geoparquet file at the end.
    """
    number_of_geometries_provided = sum(
        geom is not None
        for geom in (
            geom_filter_bbox,
            geom_filter_file,
            geom_filter_geocode,
            geom_filter_geojson,
            geom_filter_index_geohash,
            geom_filter_index_h3,
            geom_filter_index_s2,
            geom_filter_wkt,
        )
    )
    if theme_value is None or type_value is None or number_of_geometries_provided == 0:
        from click.exceptions import UsageError

        raise UsageError(
            message=(
                "OvertureMaestro requires theme, type and a geometry filter"
                " (one of --geom-filter-bbox --geom-filter-file, --geom-filter-geocode,"
                " --geom-filter-geojson, --geom-filter-index-geohash,"
                " --geom-filter-index-h3, --geom-filter-index-s2, --geom-filter-wkt)"
                " to download the data."
            ),
        )

    if number_of_geometries_provided > 1:
        raise typer.BadParameter("Provided more than one geometry for filtering")

    geometry_filter_value = (
        geom_filter_bbox
        or geom_filter_file
        or geom_filter_geocode
        or geom_filter_geojson
        or geom_filter_index_geohash
        or geom_filter_index_h3
        or geom_filter_index_s2
        or geom_filter_wkt
    )

    logging.disable(logging.CRITICAL)

    if transient_mode and silent_mode:
        raise typer.BadParameter("Cannot pass both silent and transient mode at once.")

    verbosity_mode: VERBOSITY_MODE = "verbose"

    if transient_mode:
        verbosity_mode = "transient"
    elif silent_mode:
        verbosity_mode = "silent"

    from overturemaestro import convert_geometry_to_parquet, get_available_theme_type_pairs

    if (theme_value, type_value) not in get_available_theme_type_pairs(
        release=release_version  # type: ignore[arg-type]
    ):
        raise typer.BadParameter(
            f"Dataset of theme = {theme_value} and type = {type_value} doesn't exist."
        )

    pyarrow_filter = None

    if pyarrow_filters:
        import functools
        import operator

        pyarrow_filter = functools.reduce(operator.and_, pyarrow_filters)

    geoparquet_path = convert_geometry_to_parquet(
        theme=theme_value,
        type=type_value,
        geometry_filter=geometry_filter_value,
        release=release_version,
        ignore_cache=ignore_cache,
        working_directory=working_directory,
        result_file_path=result_file_path,
        verbosity_mode=verbosity_mode,
        pyarrow_filter=pyarrow_filter,
    )

    typer.secho(geoparquet_path, fg="green")



"""Common components for tests."""

import os
import shutil
from pathlib import Path

import pytest
from pytest import Item
from shapely import Polygon, box, to_geojson, to_wkt

from overturemaestro.conftest import TEST_RELEASE_VERSIONS_LIST, patch_get_available_versions
from overturemaestro.release_index import download_existing_release_index

TEST_RELEASE_VERSION = "2024-08-20.0"

__all__ = [
    "patch_get_available_versions",
    "test_release_version",
    "TEST_RELEASE_VERSION",
    "TEST_RELEASE_VERSIONS_LIST",
    "bbox",
    "geometry_bbox_str",
    "geometry_box",
    "geometry_wkt",
    "geometry_geojson",
    "geometry_boundary_file_path",
]


def pytest_runtest_setup(item: Item) -> None:
    """Setup python encoding before `pytest_runtest_call(item)`."""
    os.environ["PYTHONIOENCODING"] = "utf-8"


@pytest.fixture(scope="session")  # type: ignore
def test_release_version() -> str:
    """Release version for tests."""
    return TEST_RELEASE_VERSION


@pytest.fixture(autouse=True, scope="session")  # type: ignore
def download_release_index(test_release_version: str) -> None:
    """Download release index for testing purposes."""
    download_existing_release_index(release=test_release_version)


@pytest.fixture(autouse=True, scope="session")  # type: ignore
def copy_geocode_cache() -> None:
    """Load cached geocoding results."""
    existing_cache_directory = Path(__file__).parent / "test_files" / "geocoding_cache"
    geocoding_cache_directory = Path("cache")
    geocoding_cache_directory.mkdir(exist_ok=True)
    for file_path in existing_cache_directory.glob("*.json"):
        destination_path = geocoding_cache_directory / file_path.name
        shutil.copy(file_path, destination_path)


def bbox() -> tuple[float, float, float, float]:
    """Bounding Box."""
    return (7.416486207767861, 43.7310867041912, 7.421931388477276, 43.73370705597216)


def geometry_bbox_str() -> str:
    """Bounding Box in str form."""
    return ",".join(map(str, bbox()))


def geometry_box() -> Polygon:
    """Geometry box."""
    return box(*bbox())


def geometry_wkt() -> str:
    """Geometry box in WKT form."""
    return str(to_wkt(geometry_box()))


def geometry_geojson() -> str:
    """Geometry box in GeoJSON form."""
    return str(to_geojson(geometry_box()))


def geometry_boundary_file_path() -> str:
    """Geometry Monaco boundary file path."""
    return str(Path(__file__).parent / "test_files" / "monaco_boundary.geojson")



"""Tests for CLI."""

import uuid
from pathlib import Path
from typing import Optional

import pytest
from parametrization import Parametrization as P
from typer.testing import CliRunner

from overturemaestro import __app_name__, __version__, cli
from overturemaestro.conftest import TEST_RELEASE_VERSIONS_LIST
from overturemaestro.data_downloader import _generate_geometry_hash
from overturemaestro.release_index import get_available_theme_type_pairs, get_newest_release_version
from tests.conftest import (
    TEST_RELEASE_VERSION,
    geometry_bbox_str,
    geometry_boundary_file_path,
    geometry_box,
    geometry_geojson,
    geometry_wkt,
)

runner = CliRunner()

# TODO: add pyarrow nested test


def random_str() -> str:
    """Return random string."""
    return str(uuid.uuid4())


def test_version() -> None:
    """Test if version is properly returned."""
    result = runner.invoke(cli.app, ["--version"])

    assert result.exit_code == 0
    assert f"{__app_name__} {__version__}\n" in result.stdout


@P.parameters("args")  # type: ignore
@P.case(
    "No args",
    [],
)  # type: ignore
@P.case(
    "Theme only",
    ["base"],
)  # type: ignore
@P.case(
    "Theme type only",
    ["base", "water"],
)  # type: ignore
def test_theme_type_and_geometry_filter_is_required(args) -> None:
    """Test if cannot run without pbf file and without geometry filter."""
    result = runner.invoke(cli.app, args)

    assert result.exit_code == 2
    assert "OvertureMaestro requires theme, type and a geometry filter" in result.stdout


def test_basic_run(test_release_version: str) -> None:
    """Test if runs properly without options."""
    theme_value = "buildings"
    type_value = "building"

    result = runner.invoke(
        cli.app,
        [
            "--release",
            test_release_version,
            "--geom-filter-bbox",
            geometry_bbox_str(),
            theme_value,
            type_value,
        ],
    )

    geometry_hash = _generate_geometry_hash(geometry_box())

    assert result.exit_code == 0
    assert (
        str(
            Path(f"files/{test_release_version}/theme={theme_value}/type={type_value}")
            / f"{geometry_hash}_nofilter.parquet"
        )
        in result.stdout
    )


def test_silent_mode(test_release_version: str) -> None:
    """Test if runs properly without reporting status."""
    theme_value = "buildings"
    type_value = "building"
    result = runner.invoke(
        cli.app,
        [
            "--release",
            test_release_version,
            "--geom-filter-bbox",
            geometry_bbox_str(),
            theme_value,
            type_value,
            "--silent",
            "--ignore-cache",
        ],
    )

    geometry_hash = _generate_geometry_hash(geometry_box())

    print(result.stdout)

    assert result.exit_code == 0
    assert (
        str(
            Path(f"files/{test_release_version}/theme={theme_value}/type={type_value}")
            / f"{geometry_hash}_nofilter.parquet"
        )
        == result.stdout.strip()
    )


def test_transient_mode(test_release_version: str) -> None:
    """Test if runs properly without reporting status."""
    theme_value = "buildings"
    type_value = "building"
    result = runner.invoke(
        cli.app,
        [
            "--release",
            test_release_version,
            "--geom-filter-bbox",
            geometry_bbox_str(),
            theme_value,
            type_value,
            "--transient",
            "--ignore-cache",
        ],
    )
    output_lines = result.stdout.strip().split("\n")

    geometry_hash = _generate_geometry_hash(geometry_box())

    assert result.exit_code == 0
    assert len(result.stdout.strip().split("\n")) == 2
    assert "Finished operation in" in output_lines[0]
    assert (
        str(
            Path(f"files/{test_release_version}/theme={theme_value}/type={type_value}")
            / f"{geometry_hash}_nofilter.parquet"
        )
        == output_lines[1]
    )


@P.parameters("args", "expected_result")  # type: ignore
@P.case(
    "Output",
    [
        "--output",
        "files/monaco_output_long.parquet",
        "--release",
        TEST_RELEASE_VERSION,
        "--geom-filter-bbox",
        geometry_bbox_str(),
        "buildings",
        "building",
    ],
    "files/monaco_output_long.parquet",
)  # type: ignore
@P.case(
    "Output short",
    [
        "-o",
        "files/monaco_output_short.parquet",
        "--release",
        TEST_RELEASE_VERSION,
        "--geom-filter-bbox",
        geometry_bbox_str(),
        "buildings",
        "building",
    ],
    "files/monaco_output_short.parquet",
)  # type: ignore
@P.case(
    "Base without release",
    [
        "--geom-filter-bbox",
        geometry_bbox_str(),
        "buildings",
        "building",
    ],
    f"files/{TEST_RELEASE_VERSIONS_LIST[-1]}/"
    f"theme=buildings/type=building/{_generate_geometry_hash(geometry_box())}_nofilter.parquet",
)  # type: ignore
@P.case(
    "Working directory",
    [
        "--release",
        TEST_RELEASE_VERSION,
        "--geom-filter-bbox",
        geometry_bbox_str(),
        "buildings",
        "building",
        "--working-directory",
        "files/workdir",
    ],
    f"files/workdir/{TEST_RELEASE_VERSION}/"
    f"theme=buildings/type=building/{_generate_geometry_hash(geometry_box())}_nofilter.parquet",
)  # type: ignore
@P.case(
    "Ignore cache",
    [
        "--release",
        TEST_RELEASE_VERSION,
        "--geom-filter-bbox",
        geometry_bbox_str(),
        "buildings",
        "building",
        "--output",
        "files/monaco_output.parquet",
        "--ignore-cache",
    ],
    "files/monaco_output.parquet",
)  # type: ignore
@P.case(
    "Ignore cache short",
    [
        "--release",
        TEST_RELEASE_VERSION,
        "--geom-filter-bbox",
        geometry_bbox_str(),
        "buildings",
        "building",
        "--output",
        "files/monaco_output.parquet",
        "--no-cache",
    ],
    "files/monaco_output.parquet",
)  # type: ignore
@P.case(
    "Silent",
    [
        "--release",
        TEST_RELEASE_VERSION,
        "--geom-filter-bbox",
        geometry_bbox_str(),
        "buildings",
        "building",
        "--silent",
        "--output",
        "files/monaco_output.parquet",
    ],
    "files/monaco_output.parquet",
)  # type: ignore
@P.case(
    "Transient",
    [
        "--release",
        TEST_RELEASE_VERSION,
        "--geom-filter-bbox",
        geometry_bbox_str(),
        "buildings",
        "building",
        "--transient",
        "--output",
        "files/monaco_output.parquet",
    ],
    "files/monaco_output.parquet",
)  # type: ignore
@P.case(
    "Output with working directory",
    [
        "--release",
        TEST_RELEASE_VERSION,
        "--geom-filter-bbox",
        geometry_bbox_str(),
        "buildings",
        "building",
        "--working-directory",
        "files/workdir",
        "-o",
        "files/monaco_output.parquet",
    ],
    "files/monaco_output.parquet",
)  # type: ignore
@P.case(
    "PyArrow filtering",
    [
        "--release",
        TEST_RELEASE_VERSION,
        "--geom-filter-bbox",
        geometry_bbox_str(),
        "buildings",
        "building",
        "--filter",
        "subtype = residential",
    ],
    f"files/{TEST_RELEASE_VERSION}/theme=buildings/type=building/"
    f"{_generate_geometry_hash(geometry_box())}_b22759b5.parquet",
)  # type: ignore
@P.case(
    "Geometry WKT filter",
    [
        "--release",
        TEST_RELEASE_VERSION,
        "--geom-filter-wkt",
        geometry_wkt(),
        "buildings",
        "building",
    ],
    f"files/{TEST_RELEASE_VERSION}/"
    "theme=buildings/type=building/09c3fc04_nofilter.parquet",
)  # type: ignore
@P.case(
    "Geometry GeoJSON filter",
    [
        "--release",
        TEST_RELEASE_VERSION,
        "--geom-filter-geojson",
        geometry_geojson(),
        "buildings",
        "building",
    ],
    f"files/{TEST_RELEASE_VERSION}/"
    "theme=buildings/type=building/82c0fdfa_nofilter.parquet",
)  # type: ignore
@P.case(
    "Geometry file filter",
    [
        "--release",
        TEST_RELEASE_VERSION,
        "--geom-filter-file",
        geometry_boundary_file_path(),
        "buildings",
        "building",
    ],
    f"files/{TEST_RELEASE_VERSION}/"
    "theme=buildings/type=building/6a869bcf_nofilter.parquet",
)  # type: ignore
@P.case(
    "Geometry geocode filter",
    [
        "--release",
        TEST_RELEASE_VERSION,
        "--geom-filter-geocode",
        "Monaco-Ville, Monaco",
        "buildings",
        "building",
    ],
    f"files/{TEST_RELEASE_VERSION}/"
    "theme=buildings/type=building/e7f0b78a_nofilter.parquet",
)  # type: ignore
@P.case(
    "Geometry Geohash filter",
    [
        "--release",
        TEST_RELEASE_VERSION,
        "--geom-filter-index-geohash",
        "spv2bc",
        "buildings",
        "building",
    ],
    f"files/{TEST_RELEASE_VERSION}/"
    "theme=buildings/type=building/c08889e8_nofilter.parquet",
)  # type: ignore
@P.case(
    "Geometry Geohash filter multiple",
    [
        "--release",
        TEST_RELEASE_VERSION,
        "--geom-filter-index-geohash",
        "spv2bc,spv2bfr",
        "buildings",
        "building",
    ],
    f"files/{TEST_RELEASE_VERSION}/"
    "theme=buildings/type=building/1bd33e0a_nofilter.parquet",
)  # type: ignore
@P.case(
    "Geometry H3 filter",
    [
        "--release",
        TEST_RELEASE_VERSION,
        "--geom-filter-index-h3",
        "8a3969a40ac7fff",
        "buildings",
        "building",
    ],
    f"files/{TEST_RELEASE_VERSION}/"
    "theme=buildings/type=building/a2f8d511_nofilter.parquet",
)  # type: ignore
@P.case(
    "Geometry H3 filter multiple",
    [
        "--release",
        TEST_RELEASE_VERSION,
        "--geom-filter-index-h3",
        "8a3969a40ac7fff,893969a4037ffff",
        "buildings",
        "building",
    ],
    f"files/{TEST_RELEASE_VERSION}/"
    "theme=buildings/type=building/e50e6489_nofilter.parquet",
)  # type: ignore
@P.case(
    "Geometry S2 filter",
    [
        "--release",
        TEST_RELEASE_VERSION,
        "--geom-filter-index-s2",
        "12cdc28bc",
        "buildings",
        "building",
    ],
    f"files/{TEST_RELEASE_VERSION}/"
    "theme=buildings/type=building/5c3d61eb_nofilter.parquet",
)  # type: ignore
@P.case(
    "Geometry S2 filter multiple",
    [
        "--release",
        TEST_RELEASE_VERSION,
        "--geom-filter-index-s2",
        "12cdc28bc,12cdc28f",
        "buildings",
        "building",
    ],
    f"files/{TEST_RELEASE_VERSION}/"
    "theme=buildings/type=building/cda5d65e_nofilter.parquet",
)  # type: ignore
def test_proper_args(args: list[str], expected_result: str) -> None:
    """Test if runs properly with options."""
    result = runner.invoke(cli.app, args)
    print(result.stdout)
    if result.exception:
        print(result.exception)

    assert result.exit_code == 0
    assert str(Path(expected_result)) in result.stdout


@P.parameters("args")  # type: ignore
@P.case(
    "Geometry WKT filter with GeoJSON",
    ["buildings", "building", "--geom-filter-wkt", geometry_geojson()],
)  # type: ignore
@P.case(
    "Geometry GeoJSON filter with WKT",
    ["buildings", "building", "--geom-filter-geojson", geometry_wkt()],
)  # type: ignore
@P.case(
    "Geometry Geohash filter with random string",
    ["buildings", "building", "--geom-filter-index-geohash", random_str()],
)  # type: ignore
@P.case(
    "Geometry H3 filter with Geohash", ["buildings", "building", "--geom-filter-index-h3", "spv2bc"]
)  # type: ignore
@P.case(
    "Geometry H3 filter with S2", ["buildings", "building", "--geom-filter-index-h3", "12cdc28bc"]
)  # type: ignore
@P.case(
    "Geometry H3 filter with random string",
    ["buildings", "building", "--geom-filter-index-h3", random_str()],
)  # type: ignore
@P.case(
    "Geometry S2 filter with random string",
    ["buildings", "building", "--geom-filter-index-s2", random_str()],
)  # type: ignore
@P.case(
    "Geometry BBOX filter with wrong values",
    ["buildings", "building", "--geom-filter-bbox", random_str()],
)  # type: ignore
@P.case(
    "Geometry BBOX filter with spaces",
    ["buildings", "building", "--geom-filter-bbox", "10,", "-5,", "6,", "17"],
)  # type: ignore
@P.case(
    "Geometry two filters",
    [
        "buildings",
        "building",
        "--geom-filter-geojson",
        geometry_geojson(),
        "--geom-filter-wkt",
        geometry_wkt(),
    ],
)  # type: ignore
@P.case(
    "Geometry nonexistent file filter",
    ["buildings", "building", "--geom-filter-file", "nonexistent_geojson_file.geojson"],
)  # type: ignore
@P.case(
    "Wrong theme & type",
    ["wrong_theme", "wrong_type", "--geom-filter-bbox", geometry_bbox_str()],
)  # type: ignore
@P.case(
    "Wrong theme",
    ["wrong_theme", "building", "--geom-filter-bbox", geometry_bbox_str()],
)  # type: ignore
@P.case(
    "Wrong type",
    ["buildings", "wrong_type", "--geom-filter-bbox", geometry_bbox_str()],
)  # type: ignore
@P.case(
    "Wrong release",
    [
        "buildings",
        "building",
        "--geom-filter-bbox",
        geometry_bbox_str(),
        "--release",
        random_str(),
    ],
)  # type: ignore
@P.case(
    "PyArrow filter with random strings",
    [
        "--release",
        TEST_RELEASE_VERSION,
        "--geom-filter-bbox",
        geometry_bbox_str(),
        "buildings",
        "building",
        "--filter",
        f"{random_str()} = {random_str()}",
    ],
)  # type: ignore
@P.case(
    "PyArrow filter with wrong columns filter",
    [
        "--release",
        TEST_RELEASE_VERSION,
        "--geom-filter-bbox",
        geometry_bbox_str(),
        "buildings",
        "building",
        "--filter",
        "confidence > 0.9",
    ],
)  # type: ignore
def test_wrong_args(args: list[str], capsys: pytest.CaptureFixture) -> None:
    """Test if doesn't run properly with options."""
    # Fix for the I/O error from the Click repository
    # https://github.com/pallets/click/issues/824#issuecomment-1583293065
    with capsys.disabled():
        result = runner.invoke(cli.app, args)
        assert result.exit_code != 0


def test_displaying_release_versions(
    capsys: pytest.CaptureFixture,
) -> None:
    """Test if displaying OSM extracts works."""
    with capsys.disabled():
        result = runner.invoke(cli.app, ["--show-release-versions"])
        output = result.stdout

        assert result.exit_code == 0
        assert len(output) > 0
        assert "Release Version" in output
        assert TEST_RELEASE_VERSION in output


@pytest.mark.parametrize(
    "release_version",
    [TEST_RELEASE_VERSION, None],
)  # type: ignore
def test_displaying_theme_type_pairs(
    release_version: Optional[str],
    capsys: pytest.CaptureFixture,
) -> None:
    """Test if displaying OSM extracts works."""
    with capsys.disabled():
        release_args = ["--release", release_version] if release_version else []
        result = runner.invoke(cli.app, ["--show-theme-type-pairs", *release_args])
        output = result.stdout

        assert result.exit_code == 0
        assert len(output) > 0
        assert (release_version or get_newest_release_version()) in output
        assert "Theme" in output
        assert "Type" in output

        if release_version:
            for theme_value, type_value in get_available_theme_type_pairs(release_version):
                assert theme_value in output
                assert type_value in output



"""Tests related to using pregenerated release indexes."""

from contextlib import nullcontext as does_not_raise
from typing import Optional

import pytest
from parametrization import Parametrization as P
from shapely import box
from shapely.geometry.base import BaseGeometry

from overturemaestro.release_index import (
    MINIMAL_SUPPORTED_RELEASE_VERSION,
    ReleaseVersionNotSupportedError,
    _check_release_version,
    get_available_theme_type_pairs,
    load_release_index,
    load_release_indexes,
)
from tests.conftest import TEST_RELEASE_VERSION


@pytest.mark.parametrize(
    "expectation,release_version",
    [
        (pytest.raises(ReleaseVersionNotSupportedError), "2024-03-12-alpha.0"),
        (does_not_raise(), MINIMAL_SUPPORTED_RELEASE_VERSION),
        (does_not_raise(), TEST_RELEASE_VERSION),
    ],
)  # type: ignore
def test_supported_release_version(expectation, release_version: str) -> None:
    """Test if raises errors for unsupported release versions."""
    with expectation:
        _check_release_version(release_version)


def test_get_available_theme_type_pairs(test_release_version: str) -> None:
    """Test is loading available theme type pairs works."""
    pairs = get_available_theme_type_pairs(test_release_version)
    assert len(pairs) == 14


@P.parameters("theme_type_pair", "geometry_filter", "expected_number_of_rows")  # type: ignore
@P.case(
    "Base water - no filter",
    ("base", "water"),
    None,
    74348,
)  # type: ignore
@P.case(
    "Buildings - bbox filter",
    ("buildings", "building"),
    box(-0.120077, 51.498164, -0.090809, 51.508849),
    4,
)  # type: ignore
def test_load_release_index(
    test_release_version,
    theme_type_pair: tuple[str, str],
    geometry_filter: Optional[BaseGeometry],
    expected_number_of_rows: int,
) -> None:
    """Test is load_release_index function works."""
    theme_value, type_value = theme_type_pair
    release_index = load_release_index(
        release=test_release_version,
        theme=theme_value,
        type=type_value,
        geometry_filter=geometry_filter,
    )
    assert (
        len(release_index) == expected_number_of_rows
    ), f"Mismatch in number of rows: {len(release_index)} vs {expected_number_of_rows}"
    assert (
        release_index["filename"]
        .str.startswith(
            f"overturemaps-us-west-2/release/{test_release_version}/theme={theme_value}/type={type_value}"
        )
        .all()
    ), "Some rows don't match expected filename structure"


def test_load_release_indexes(test_release_version: str) -> None:
    """Test is load_release_indexes function works."""
    theme_type_pairs = [("base", "water"), ("buildings", "building")]
    combined_release_index = load_release_indexes(
        release=test_release_version,
        theme_type_pairs=theme_type_pairs,
    )
    total_expected_length = sum(
        len(
            load_release_index(
                release=test_release_version,
                theme=theme_type_pair[0],
                type=theme_type_pair[1],
            )
        )
        for theme_type_pair in theme_type_pairs
    )
    assert len(combined_release_index) == total_expected_length



"""Test big file pipeline."""

from parametrization import Parametrization as P

from overturemaestro import convert_geometry_to_parquet, geocode_to_geometry


@P.parameters("geocode_filter", "theme_type_pair")  # type: ignore
@P.case("Spain", "spain", ("buildings", "building"))  # type: ignore
def test_big_file(
    geocode_filter: str, theme_type_pair: tuple[str, str], test_release_version: str
) -> None:
    """Test if big file is working in a low memory environment."""
    geometry_filter = geocode_to_geometry(geocode_filter)
    convert_geometry_to_parquet(
        geometry_filter=geometry_filter,
        theme=theme_type_pair[0],
        type=theme_type_pair[1],
        release=test_release_version,
        verbosity_mode="verbose",
        ignore_cache=True,
    )



"""
OvertureMaestro.

OvertureMaestro is a Python library used for downloading Overture Maps datasets with quality of life
features.
"""

from overturemaestro.functions import (
    convert_bounding_box_to_geodataframe,
    convert_bounding_box_to_geodataframe_for_multiple_types,
    convert_bounding_box_to_parquet,
    convert_bounding_box_to_parquet_for_multiple_types,
    convert_geometry_to_geodataframe,
    convert_geometry_to_geodataframe_for_multiple_types,
    convert_geometry_to_parquet,
    convert_geometry_to_parquet_for_multiple_types,
)
from overturemaestro.geocode import geocode_to_geometry
from overturemaestro.release_index import (
    get_available_release_versions,
    get_available_theme_type_pairs,
    get_newest_release_version,
)

__app_name__ = "OvertureMaestro"
__version__ = "0.2.6"

__all__ = [
    "convert_bounding_box_to_geodataframe",
    "convert_bounding_box_to_geodataframe_for_multiple_types",
    "convert_bounding_box_to_parquet",
    "convert_bounding_box_to_parquet_for_multiple_types",
    "convert_geometry_to_geodataframe",
    "convert_geometry_to_geodataframe_for_multiple_types",
    "convert_geometry_to_parquet",
    "convert_geometry_to_parquet_for_multiple_types",
    "geocode_to_geometry",
    "get_available_release_versions",
    "get_available_theme_type_pairs",
    "get_newest_release_version",
]



"""Decorator for showing total elapsed time at the end of operation."""

from typing import Any, Callable, TypeVar, cast

F = TypeVar("F", bound=Callable[..., Any])


def show_total_elapsed_time_decorator(func: F) -> F:
    """Decorator for showing total elapsed time at the end of operation."""
    import time
    from functools import wraps

    from overturemaestro._rich_progress import show_total_elapsed_time

    @wraps(func)
    def timeit_wrapper(*args: Any, **kwargs: Any) -> Any:
        start_time = time.time()

        result = func(*args, **kwargs)

        verbosity_mode = kwargs.get("verbosity_mode", "silent")
        if not is_inside_decorated_calling_function() and not verbosity_mode == "silent":
            end_time = time.time()
            elapsed_seconds = end_time - start_time
            show_total_elapsed_time(elapsed_seconds)

        return result

    func.__is_timed__ = True  # type: ignore[attr-defined]

    return cast(F, timeit_wrapper)


def is_decorated(func: F) -> bool:
    """Get information if functions is already timed."""
    return hasattr(func, "__is_timed__")


# https://stackoverflow.com/a/39079070
def is_inside_decorated_calling_function() -> bool:
    """Finds if any of the calling functions is decorated."""
    import inspect

    for stack_frame in inspect.stack()[1:]:
        fr = stack_frame[0]
        co = fr.f_code
        for get in (
            lambda fr=fr, co=co: fr.f_globals[co.co_name],
            lambda fr=fr, co=co: getattr(fr.f_locals["self"], co.co_name),
            lambda fr=fr, co=co: getattr(fr.f_locals["cls"], co.co_name),
            lambda fr=fr, co=co: fr.f_back.f_locals[co.co_name],  # nested
            lambda fr=fr: fr.f_back.f_locals["func"],  # decorators
            lambda fr=fr: fr.f_back.f_locals["meth"],
            lambda fr=fr: fr.f_back.f_locals["f"],
        ):
            try:
                func = get()  # type: ignore[no-untyped-call]
            except (KeyError, AttributeError):
                pass
            else:
                if func is not None and hasattr(func, "__code__") and func.__code__ == co and is_decorated(func):
                    return True

    return False



import itertools
from collections.abc import Generator, Iterable
from typing import Any

EARTH_RADIUS_KM = 6371
CLUSTERING_THRESHOLD = 20 / EARTH_RADIUS_KM  # calculate 20 kilometers threshold


def calculate_row_group_bounding_box(
    parquet_filename: str, parquet_row_group: int, pyarrow_table: Any
) -> Any:
    import polars as pl
    from sklearn.cluster import Birch

    df = (
        pl.from_arrow(pyarrow_table)
        .with_row_index(name="row_index")
        .unnest("bbox")
        .select(
            [
                pl.col("row_index"),
                pl.col("xmin"),
                pl.col("xmax"),
                pl.col("ymin"),
                pl.col("ymax"),
                pl.sum_horizontal("xmin", "xmax").truediv(2).radians().alias("x_rad"),
                pl.sum_horizontal("ymin", "ymax").truediv(2).radians().alias("y_rad"),
            ]
        )
    )

    cluster_labels = (
        Birch(n_clusters=None, threshold=CLUSTERING_THRESHOLD).fit(df[["x_rad", "y_rad"]]).labels_
    )

    return (
        df.with_columns(pl.Series(name="label", values=cluster_labels))
        .group_by("label")
        .agg(
            pl.col("xmin").min(),
            pl.col("xmax").max(),
            pl.col("ymin").min(),
            pl.col("ymax").max(),
            pl.col("row_index").alias("row_indexes"),
        )
        .select(
            pl.lit(parquet_filename).alias("filename"),
            pl.lit(parquet_row_group).alias("row_group"),
            pl.col("xmin"),
            pl.col("xmax"),
            pl.col("ymin"),
            pl.col("ymax"),
            pl.col("row_indexes")
            .map_elements(
                lambda lst: list(calculate_ranges(lst)), return_dtype=pl.List(pl.List(pl.Int64))
            )
            .cast(pl.List(pl.List(pl.UInt32)))
            .alias("row_indexes_ranges"),
        )
    ).to_arrow()


# https://stackoverflow.com/a/4629241
def calculate_ranges(i: Iterable[int]) -> Generator[tuple[int, int]]:
    for _a, b in itertools.groupby(enumerate(i), lambda pair: pair[1] - pair[0]):
        b_lst = list(b)
        yield b_lst[0][1], b_lst[-1][1]


def decompress_ranges(ranges: list[list[int]]) -> list[int]:
    return list(
        itertools.chain.from_iterable(
            range(range_bounds[0], range_bounds[1] + 1) for range_bounds in ranges
        )
    )



"""
Data downloader utilities.

Functions used to download Overture Maps data before local filtering.
"""

import multiprocessing
import operator
from pathlib import Path
from typing import TYPE_CHECKING, Any, Optional, Union, cast, overload

from overturemaestro._exceptions import MissingColumnError
from overturemaestro.elapsed_time_decorator import show_total_elapsed_time_decorator

if TYPE_CHECKING:  # pragma: no cover
    from pyarrow import Schema
    from pyarrow.compute import Expression
    from shapely.geometry.base import BaseGeometry

    from overturemaestro._rich_progress import VERBOSITY_MODE

PYARROW_EXPRESSION = tuple[Any, Any, Any]
PYARROW_FILTER = Union["Expression", list[PYARROW_EXPRESSION], list[list[PYARROW_EXPRESSION]]]

__all__ = [
    "download_data",
    "download_data_for_multiple_types",
]


@overload
def download_data_for_multiple_types(
    theme_type_pairs: list[tuple[str, str]],
    geometry_filter: "BaseGeometry",
    *,
    pyarrow_filters: Optional[list[Optional[PYARROW_FILTER]]],
    columns_to_download: Optional[list[Optional[list[str]]]],
    ignore_cache: bool = False,
    working_directory: Union[str, Path] = "files",
    verbosity_mode: "VERBOSITY_MODE" = "transient",
    max_workers: Optional[int] = None,
) -> list[Path]: ...


@overload
def download_data_for_multiple_types(
    theme_type_pairs: list[tuple[str, str]],
    geometry_filter: "BaseGeometry",
    release: str,
    *,
    pyarrow_filters: Optional[list[Optional[PYARROW_FILTER]]],
    columns_to_download: Optional[list[Optional[list[str]]]],
    ignore_cache: bool = False,
    working_directory: Union[str, Path] = "files",
    verbosity_mode: "VERBOSITY_MODE" = "transient",
    max_workers: Optional[int] = None,
) -> list[Path]: ...


@overload
def download_data_for_multiple_types(
    theme_type_pairs: list[tuple[str, str]],
    geometry_filter: "BaseGeometry",
    release: Optional[str] = None,
    *,
    pyarrow_filters: Optional[list[Optional[PYARROW_FILTER]]],
    columns_to_download: Optional[list[Optional[list[str]]]],
    ignore_cache: bool = False,
    working_directory: Union[str, Path] = "files",
    verbosity_mode: "VERBOSITY_MODE" = "transient",
    max_workers: Optional[int] = None,
) -> list[Path]: ...


@show_total_elapsed_time_decorator
def download_data_for_multiple_types(
    theme_type_pairs: list[tuple[str, str]],
    geometry_filter: "BaseGeometry",
    release: Optional[str] = None,
    *,
    pyarrow_filters: Optional[list[Optional[PYARROW_FILTER]]] = None,
    columns_to_download: Optional[list[Optional[list[str]]]] = None,
    ignore_cache: bool = False,
    working_directory: Union[str, Path] = "files",
    verbosity_mode: "VERBOSITY_MODE" = "transient",
    max_workers: Optional[int] = None,
) -> list[Path]:
    """
    Downloads the data for the given release for multiple types.

    Args:
        theme_type_pairs (list[tuple[str, str]]): Pairs of themes and types of the dataset.
        geometry_filter (BaseGeometry): Geometry used to filter data.
        release (Optional[str], optional): Release version. If not provided, will automatically load
            newest available release version. Defaults to None.
        pyarrow_filters (Optional[list[Optional[PYARROW_FILTER]]], optional): A list of pyarrow
            expressions used to filter specific theme type pair. Must be the same length as the list
            of theme type pairs. Defaults to None.
        columns_to_download (Optional[list[Optional[list[str]]]], optional): A list of pyarrow
            expressions used to filter specific theme type pair. Must be the same length as the list
            of theme type pairs. Defaults to None.
        ignore_cache (bool, optional): Whether to ignore precalculated geoparquet files or not.
            Defaults to False.
        working_directory (Union[str, Path], optional): Directory where to save
            the downloaded `*.parquet` files. Defaults to "files".
        verbosity_mode (Literal["silent", "transient", "verbose"], optional): Set progress
            verbosity mode. Can be one of: silent, transient and verbose. Silent disables
            output completely. Transient tracks progress, but removes output after finished.
            Verbose leaves all progress outputs in the stdout. Defaults to "transient".
        max_workers (Optional[int], optional): Max number of multiprocessing workers used to
            process the dataset. Defaults to None.

    Returns:
        list[Path]: List of saved Geoparquet files paths.
    """
    import tempfile

    import pyarrow.dataset as ds
    import pyarrow.parquet as pq

    from overturemaestro._rich_progress import TrackProgressBar
    from overturemaestro.release_index import get_newest_release_version

    if pyarrow_filters is not None and len(theme_type_pairs) != len(pyarrow_filters):
        raise ValueError("Pyarrow filters length doesn't match length of theme type pairs.")

    if columns_to_download is not None and len(theme_type_pairs) != len(columns_to_download):
        raise ValueError("Columns to download length doesn't match length of theme type pairs.")

    if not release:
        release = get_newest_release_version()

    working_directory = Path(working_directory)
    working_directory.mkdir(parents=True, exist_ok=True)

    all_result_file_paths = []
    theme_type_pairs_to_download = []
    pyarrow_filters_list = []
    columns_to_download_list = []
    result_file_paths_to_download = []
    for idx, (theme_value, type_value) in enumerate(theme_type_pairs):
        _pyarrow_filter = pyarrow_filters[idx] if pyarrow_filters else None
        _columns_to_download = columns_to_download[idx] if columns_to_download else None

        if _pyarrow_filter is not None:
            from pyarrow.parquet import filters_to_expression

            _pyarrow_filter = filters_to_expression(_pyarrow_filter)

        result_file_path = working_directory / _generate_result_file_path(
            release=release,
            theme=theme_value,
            type=type_value,
            geometry_filter=geometry_filter,
            pyarrow_filter=_pyarrow_filter,
            columns_to_download=_columns_to_download,
        )
        all_result_file_paths.append(result_file_path)

        if not result_file_path.exists() or ignore_cache:
            theme_type_pairs_to_download.append((theme_value, type_value))
            pyarrow_filters_list.append(_pyarrow_filter)
            columns_to_download_list.append(_columns_to_download)
            result_file_paths_to_download.append(result_file_path)

    if theme_type_pairs_to_download:
        with tempfile.TemporaryDirectory(dir=Path(working_directory).resolve()) as tmp_dir_name:
            tmp_dir_path = Path(tmp_dir_name)
            raw_parquet_files_per_pair = _download_data(
                release=release,
                theme_type_pairs=theme_type_pairs_to_download,
                geometry_filter=geometry_filter,
                pyarrow_filters=pyarrow_filters_list,
                columns_to_download=columns_to_download_list,
                working_directory=tmp_dir_path,
                verbosity_mode=verbosity_mode,
                max_workers=max_workers,
            )

            with TrackProgressBar(verbosity_mode=verbosity_mode) as progress:
                for result_file_path, raw_parquet_files in progress.track(
                    zip(result_file_paths_to_download, raw_parquet_files_per_pair),
                    total=len(result_file_paths_to_download),
                    description="Saving final geoparquet files",
                ):
                    final_dataset = ds.dataset(raw_parquet_files)
                    result_file_path.parent.mkdir(exist_ok=True, parents=True)
                    with pq.ParquetWriter(result_file_path, final_dataset.schema) as writer:
                        for batch in final_dataset.to_batches():
                            writer.write_batch(batch)

    return all_result_file_paths


@overload
def download_data(
    theme: str,
    type: str,
    geometry_filter: "BaseGeometry",
    *,
    pyarrow_filter: Optional[PYARROW_FILTER] = None,
    columns_to_download: Optional[list[str]] = None,
    result_file_path: Optional[Union[str, Path]] = None,
    ignore_cache: bool = False,
    working_directory: Union[str, Path] = "files",
    verbosity_mode: "VERBOSITY_MODE" = "transient",
    max_workers: Optional[int] = None,
) -> Path: ...


@overload
def download_data(
    theme: str,
    type: str,
    geometry_filter: "BaseGeometry",
    release: str,
    *,
    pyarrow_filter: Optional[PYARROW_FILTER] = None,
    columns_to_download: Optional[list[str]] = None,
    result_file_path: Optional[Union[str, Path]] = None,
    ignore_cache: bool = False,
    working_directory: Union[str, Path] = "files",
    verbosity_mode: "VERBOSITY_MODE" = "transient",
    max_workers: Optional[int] = None,
) -> Path: ...


@overload
def download_data(
    theme: str,
    type: str,
    geometry_filter: "BaseGeometry",
    release: Optional[str] = None,
    *,
    pyarrow_filter: Optional[PYARROW_FILTER] = None,
    columns_to_download: Optional[list[str]] = None,
    result_file_path: Optional[Union[str, Path]] = None,
    ignore_cache: bool = False,
    working_directory: Union[str, Path] = "files",
    verbosity_mode: "VERBOSITY_MODE" = "transient",
    max_workers: Optional[int] = None,
) -> Path: ...


@show_total_elapsed_time_decorator
def download_data(
    theme: str,
    type: str,
    geometry_filter: "BaseGeometry",
    release: Optional[str] = None,
    *,
    pyarrow_filter: Optional[PYARROW_FILTER] = None,
    columns_to_download: Optional[list[str]] = None,
    result_file_path: Optional[Union[str, Path]] = None,
    ignore_cache: bool = False,
    working_directory: Union[str, Path] = "files",
    verbosity_mode: "VERBOSITY_MODE" = "transient",
    max_workers: Optional[int] = None,
) -> Path:
    """
    Downloads the data for the given release.

    Args:
        theme (str): Theme of the dataset.
        type (str): Type of the dataset.
        geometry_filter (BaseGeometry): Geometry used to filter data.
        release (Optional[str], optional): Release version. If not provided, will automatically load
            newest available release version. Defaults to None.
        pyarrow_filter (Optional[PYARROW_FILTER], optional): Filters to apply on a pyarrow dataset.
            Can be pyarrow.compute.Expression or List[Tuple] or List[List[Tuple]]. Defaults to None.
        columns_to_download (Optional[list[str]], optional): List of columns to download.
            Automatically adds geometry column to the list. If None, will download all columns.
            Defaults to None.
        result_file_path (Union[str, Path], optional): Where to save
            the geoparquet file. If not provided, will be generated based on hashes
            from filters. Defaults to None.
        ignore_cache (bool, optional): Whether to ignore precalculated geoparquet files or not.
            Defaults to False.
        working_directory (Union[str, Path], optional): Directory where to save
            the downloaded `*.parquet` files. Defaults to "files".
        verbosity_mode (Literal["silent", "transient", "verbose"], optional): Set progress
            verbosity mode. Can be one of: silent, transient and verbose. Silent disables
            output completely. Transient tracks progress, but removes output after finished.
            Verbose leaves all progress outputs in the stdout. Defaults to "transient".
        max_workers (Optional[int], optional): Max number of multiprocessing workers used to
            process the dataset. Defaults to None.

    Returns:
        Path: Saved Geoparquet file path.
    """
    import tempfile

    import pyarrow.dataset as ds
    import pyarrow.parquet as pq

    from overturemaestro._rich_progress import TrackProgressSpinner
    from overturemaestro.release_index import get_newest_release_version

    if not release:
        release = get_newest_release_version()

    working_directory = Path(working_directory)
    working_directory.mkdir(parents=True, exist_ok=True)

    if pyarrow_filter is not None:
        from pyarrow.parquet import filters_to_expression

        pyarrow_filter = filters_to_expression(pyarrow_filter)

    if result_file_path is None:
        result_file_path = working_directory / _generate_result_file_path(
            release=release,
            theme=theme,
            type=type,
            geometry_filter=geometry_filter,
            pyarrow_filter=pyarrow_filter,
            columns_to_download=columns_to_download,
        )

    result_file_path = Path(result_file_path)

    if not result_file_path.exists() or ignore_cache:
        with tempfile.TemporaryDirectory(dir=Path(working_directory).resolve()) as tmp_dir_name:
            tmp_dir_path = Path(tmp_dir_name)
            raw_parquet_files = _download_data(
                release=release,
                theme_type_pairs=[(theme, type)],
                geometry_filter=geometry_filter,
                pyarrow_filters=[pyarrow_filter],
                columns_to_download=[columns_to_download],
                working_directory=tmp_dir_path,
                verbosity_mode=verbosity_mode,
                max_workers=max_workers,
            )[0]

            final_dataset = ds.dataset(raw_parquet_files)

            result_file_path.parent.mkdir(exist_ok=True, parents=True)

            with (
                TrackProgressSpinner("Saving final geoparquet file", verbosity_mode=verbosity_mode),
                pq.ParquetWriter(result_file_path, final_dataset.schema) as writer,
            ):
                for batch in final_dataset.to_batches():
                    writer.write_batch(batch)

    return result_file_path


@show_total_elapsed_time_decorator
def _download_data(
    release: str,
    theme_type_pairs: list[tuple[str, str]],
    geometry_filter: "BaseGeometry",
    pyarrow_filters: Optional[list[Union["Expression", None]]],
    columns_to_download: Optional[list[Optional[list[str]]]],
    working_directory: Path,
    verbosity_mode: "VERBOSITY_MODE",
    max_workers: Optional[int],
) -> list[list[Path]]:
    if pyarrow_filters is not None and len(theme_type_pairs) != len(pyarrow_filters):
        raise ValueError("Pyarrow filters length doesn't match length of theme type pairs.")

    if columns_to_download is not None and len(theme_type_pairs) != len(columns_to_download):
        raise ValueError("Columns to download length doesn't match length of theme type pairs.")

    # force tuple structure
    theme_type_pairs = [(theme_value, type_value) for theme_value, type_value in theme_type_pairs]

    all_row_groups_to_download = _prepare_row_groups_for_download(
        release=release,
        theme_type_pairs=theme_type_pairs,
        geometry_filter=geometry_filter,
        pyarrow_filters=pyarrow_filters,
        columns_to_download=columns_to_download,
        verbosity_mode=verbosity_mode,
    )

    downloaded_parquet_files = _download_prepared_row_groups(
        all_row_groups_to_download=all_row_groups_to_download,
        theme_type_pairs=theme_type_pairs,
        geometry_filter=geometry_filter,
        working_directory=working_directory,
        verbosity_mode=verbosity_mode,
        max_workers=max_workers,
    )

    filtered_parquet_files = _filter_downloaded_files(
        downloaded_parquet_files=downloaded_parquet_files,
        theme_type_pairs=theme_type_pairs,
        geometry_filter=geometry_filter,
        working_directory=working_directory,
        verbosity_mode=verbosity_mode,
        max_workers=max_workers,
    )

    grouped_parquet_paths = _group_parquet_files(
        filtered_parquet_files=filtered_parquet_files,
        release=release,
        theme_type_pairs=theme_type_pairs,
        columns_to_download=columns_to_download,
        working_directory=working_directory,
        verbosity_mode=verbosity_mode,
    )

    return grouped_parquet_paths


def _prepare_row_groups_for_download(
    release: str,
    theme_type_pairs: list[tuple[str, str]],
    geometry_filter: "BaseGeometry",
    pyarrow_filters: Optional[list[Union["Expression", None]]],
    columns_to_download: Optional[list[Optional[list[str]]]],
    verbosity_mode: "VERBOSITY_MODE",
) -> list[dict[str, Any]]:
    from overturemaestro.release_index import load_release_index

    all_row_groups_to_download = []

    for idx, (theme_value, type_value) in enumerate(theme_type_pairs):
        dataset_index = load_release_index(
            release=release,
            theme=theme_value,
            type=type_value,
            geometry_filter=geometry_filter,
            verbosity_mode=verbosity_mode,
        )

        dataset_index = (
            dataset_index.explode("row_indexes_ranges")
            .groupby(["filename", "row_group"])["row_indexes_ranges"]
            .apply(lambda x: [pair.tolist() for pair in x.values])
            .reset_index()
        )

        row_groups_to_download = dataset_index[
            ["filename", "row_group", "row_indexes_ranges"]
        ].to_dict(orient="records")

        _pyarrow_filter = pyarrow_filters[idx] if pyarrow_filters else None
        _columns_to_download = columns_to_download[idx] if columns_to_download else None

        for row_group_to_download in row_groups_to_download:
            row_group_to_download["theme"] = theme_value
            row_group_to_download["type"] = type_value
            row_group_to_download["user_defined_pyarrow_filter"] = _pyarrow_filter
            row_group_to_download["columns_to_download"] = _columns_to_download

        all_row_groups_to_download.extend(row_groups_to_download)

    return all_row_groups_to_download


def _download_prepared_row_groups(
    all_row_groups_to_download: list[dict[str, Any]],
    theme_type_pairs: list[tuple[str, str]],
    geometry_filter: "BaseGeometry",
    working_directory: Path,
    verbosity_mode: "VERBOSITY_MODE",
    max_workers: Optional[int],
) -> list[Path]:
    from concurrent.futures import ProcessPoolExecutor
    from functools import partial

    from overturemaestro._rich_progress import TrackProgressBar

    total_row_groups = len(all_row_groups_to_download)

    if total_row_groups == 0:
        return []

    min_no_workers = 8
    no_workers = min(
        max(min_no_workers, multiprocessing.cpu_count() + 4), 32
    )  # minimum 8 workers, but not more than 32

    if max_workers:
        no_workers = min(max_workers, no_workers)

    with (
        TrackProgressBar(verbosity_mode=verbosity_mode) as progress,
        ProcessPoolExecutor(
            max_workers=min(no_workers, total_row_groups),
            mp_context=multiprocessing.get_context("spawn"),
        ) as ex,
    ):
        fn = partial(
            _download_single_parquet_row_group_multiprocessing,
            bbox=geometry_filter.bounds,
            working_directory=working_directory,
        )
        theme_type_task_description = (
            f"{theme_type_pairs[0][0]}/{theme_type_pairs[0][1]}"
            if len(theme_type_pairs) == 1
            else f"{len(theme_type_pairs)} datasets"
        )
        downloaded_parquet_files = list(
            progress.track(
                ex.map(
                    fn,
                    all_row_groups_to_download,
                    chunksize=1,
                ),
                description=f"Downloading parquet files ({theme_type_task_description})",
                total=total_row_groups,
            )
        )

    return downloaded_parquet_files


def _filter_downloaded_files(
    downloaded_parquet_files: list[Path],
    theme_type_pairs: list[tuple[str, str]],
    geometry_filter: "BaseGeometry",
    working_directory: Path,
    verbosity_mode: "VERBOSITY_MODE",
    max_workers: Optional[int],
) -> list[Path]:
    from functools import partial

    from overturemaestro._parquet_multiprocessing import map_parquet_dataset

    if not geometry_filter.equals(geometry_filter.envelope):
        destination_path = working_directory / Path("intersected_data")
        fn = partial(_filter_data_properly, geometry_filter=geometry_filter)
        theme_type_task_description = (
            f"{theme_type_pairs[0][0]}/{theme_type_pairs[0][1]}"
            if len(theme_type_pairs) == 1
            else f"{len(theme_type_pairs)} datasets"
        )
        map_parquet_dataset(
            dataset_path=downloaded_parquet_files,
            destination_path=destination_path,
            function=fn,
            progress_description=f"Filtering data by geometry ({theme_type_task_description})",
            report_progress_as_text=False,
            verbosity_mode=verbosity_mode,
            max_workers=max_workers,
        )

        filtered_parquet_files = list(destination_path.glob("**/*.parquet"))
    else:
        filtered_parquet_files = downloaded_parquet_files

    return filtered_parquet_files


def _group_parquet_files(
    filtered_parquet_files: list[Path],
    release: str,
    theme_type_pairs: list[tuple[str, str]],
    columns_to_download: Optional[list[Optional[list[str]]]],
    working_directory: Path,
    verbosity_mode: "VERBOSITY_MODE",
) -> list[list[Path]]:
    if len(theme_type_pairs) == 1:
        grouped_parquet_paths = [filtered_parquet_files]
    else:
        grouped_parquet_paths = [[] for _ in theme_type_pairs]
        for parquet_file in filtered_parquet_files:
            import pyarrow.parquet as pq

            metadata = pq.read_schema(parquet_file).metadata
            theme_value = metadata[b"_theme"].decode()
            type_value = metadata[b"_type"].decode()
            idx = theme_type_pairs.index((theme_value, type_value))
            grouped_parquet_paths[idx].append(parquet_file)

    # Check for empty files
    for idx, (theme_value, type_value) in enumerate(theme_type_pairs):
        if not grouped_parquet_paths[idx]:
            grouped_parquet_paths[idx].append(
                _generate_empty_file(
                    theme=theme_value,
                    type=type_value,
                    release=release,
                    columns_to_download=(
                        columns_to_download[idx] if columns_to_download is not None else None
                    ),
                    working_directory=working_directory,
                    verbosity_mode=verbosity_mode,
                )
            )

    return grouped_parquet_paths


def _generate_empty_file(
    theme: str,
    type: str,
    release: str,
    columns_to_download: Optional[list[str]],
    working_directory: Path,
    verbosity_mode: "VERBOSITY_MODE",
) -> Path:
    import pyarrow as pa
    import pyarrow.fs as fs
    import pyarrow.parquet as pq
    from overturemaps.core import geoarrow_schema_adapter

    from overturemaestro._rich_progress import TrackProgressSpinner
    from overturemaestro.release_index import load_release_index

    result_file_path = working_directory / f"empty_{theme}_{type}.parquet"
    result_file_path.parent.mkdir(parents=True, exist_ok=True)

    with TrackProgressSpinner(
        f"Preparing empty file ({theme}/{type})", verbosity_mode=verbosity_mode
    ):
        index = load_release_index(release=release, theme=theme, type=type, verbosity_mode="silent")
        remote_file = index.iloc[0].filename

        schema = pq.read_schema(
            remote_file, filesystem=fs.S3FileSystem(anonymous=True, region="us-west-2")
        )

        geoarrow_full_schema = geoarrow_schema_adapter(schema)
        metadata = geoarrow_full_schema.metadata or {}
        metadata["_theme"] = theme
        metadata["_type"] = type
        geoarrow_full_schema = geoarrow_full_schema.with_metadata(metadata)
        geoarrow_schema_filtered = geoarrow_full_schema

        if columns_to_download:
            nonexistent_columns = set(columns_to_download).difference(geoarrow_full_schema.names)

            if nonexistent_columns:
                raise MissingColumnError(
                    f"Cannot download given columns: {', '.join(sorted(nonexistent_columns))}"
                )

            if "geometry" not in columns_to_download:
                columns_to_download.append("geometry")

            # Reorder list of columns to download to match pyarrow schema
            columns_to_download = [
                column_name
                for column_name in geoarrow_full_schema.names
                if column_name in columns_to_download
            ]

            # Remove unused columns from schema
            for column_name in geoarrow_schema_filtered.names:
                if column_name in columns_to_download:
                    continue
                geoarrow_schema_filtered = geoarrow_schema_filtered.remove(
                    geoarrow_schema_filtered.get_field_index(column_name)
                )

        pq.write_table(table=pa.Table.from_pylist([], schema=schema), where=result_file_path)

    return result_file_path


def _download_single_parquet_row_group_multiprocessing(
    params: dict[str, Any],
    bbox: tuple[float, float, float, float],
    working_directory: Path,
) -> Path:
    from pyarrow.lib import ArrowInvalid

    retries = 10
    while retries > 0:
        try:
            downloaded_path = _download_single_parquet_row_group(
                **params,
                bbox=bbox,
                working_directory=working_directory,
            )
            return downloaded_path
        except (MissingColumnError, ArrowInvalid):
            raise
        except Exception:  # pragma: no cover
            retries -= 1
            if retries == 0:
                raise

    raise Exception()


def _download_single_parquet_row_group(
    theme: str,
    type: str,
    filename: str,
    row_group: int,
    row_indexes_ranges: list[list[int]],
    bbox: tuple[float, float, float, float],
    user_defined_pyarrow_filter: Optional["Expression"],
    columns_to_download: Optional[list[str]],
    working_directory: Path,
) -> Path:
    import pyarrow.compute as pc
    import pyarrow.dataset as ds
    import pyarrow.fs as fs
    from overturemaps.cli import get_writer
    from overturemaps.core import geoarrow_schema_adapter

    from overturemaestro._geometry_clustering import decompress_ranges

    row_indexes = decompress_ranges(row_indexes_ranges)
    xmin, ymin, xmax, ymax = bbox
    pyarrow_filter = (
        (pc.field("bbox", "xmin") < xmax)
        & (pc.field("bbox", "xmax") > xmin)
        & (pc.field("bbox", "ymin") < ymax)
        & (pc.field("bbox", "ymax") > ymin)
    )

    if user_defined_pyarrow_filter is not None:
        pyarrow_filter = pyarrow_filter & user_defined_pyarrow_filter

    fragment_manual = ds.ParquetFileFormat().make_fragment(
        file=filename,
        filesystem=fs.S3FileSystem(anonymous=True, region="us-west-2"),
        row_groups=[row_group],
    )
    geoarrow_full_schema = geoarrow_schema_adapter(fragment_manual.physical_schema)
    metadata = geoarrow_full_schema.metadata or {}
    metadata["_theme"] = theme
    metadata["_type"] = type
    geoarrow_full_schema = geoarrow_full_schema.with_metadata(metadata)
    geoarrow_schema_filtered = geoarrow_full_schema

    filtering_columns = None

    columns_to_remove = set()

    if columns_to_download:
        nonexistent_columns = set(columns_to_download).difference(geoarrow_full_schema.names)

        if nonexistent_columns:
            raise MissingColumnError(
                f"Cannot download given columns: {', '.join(sorted(nonexistent_columns))}"
            )

        if "geometry" not in columns_to_download:
            columns_to_download.append("geometry")

        # Create list of columns used to filter data by bbox
        columns_required_for_filtering = {"bbox"}
        if user_defined_pyarrow_filter is not None:
            columns_required_for_filtering = columns_required_for_filtering.union(
                _get_filtering_columns_from_pyarrow_filter(
                    schema=geoarrow_full_schema, pyarrow_filter=user_defined_pyarrow_filter
                )
            )

        filtering_columns = [
            column_name
            for column_name in geoarrow_full_schema.names
            if column_name in columns_to_download or column_name in columns_required_for_filtering
        ]

        # Reorder list of columns to download to match pyarrow schema
        columns_to_download = [
            column_name
            for column_name in geoarrow_full_schema.names
            if column_name in columns_to_download
        ]

        # Remove unused columns from schema
        for column_name in geoarrow_schema_filtered.names:
            if column_name in columns_to_download:
                continue
            geoarrow_schema_filtered = geoarrow_schema_filtered.remove(
                geoarrow_schema_filtered.get_field_index(column_name)
            )

        columns_to_remove = set(filtering_columns).difference(columns_to_download)

    file_path = working_directory / _generate_row_group_file_name(
        filename, row_group, row_indexes_ranges, bbox
    )
    file_path.parent.mkdir(parents=True, exist_ok=True)

    with get_writer(
        output_format="geoparquet", path=file_path, schema=geoarrow_schema_filtered
    ) as writer:
        writer.write_table(
            fragment_manual.scanner(schema=geoarrow_full_schema, columns=filtering_columns)
            .take(row_indexes)
            .filter(pyarrow_filter)
            .drop_columns(list(columns_to_remove))
        )

    return file_path


def _get_filtering_columns_from_pyarrow_filter(
    schema: "Schema", pyarrow_filter: "Expression"
) -> set[str]:
    import re

    import pyarrow.substrait as pas

    parsed_expression = pas.deserialize_expressions(
        pyarrow_filter.to_substrait(schema=schema)
    ).expressions["expression"]

    pattern = r"FieldPath\((\d+(?: \d+)*)\)"
    field_paths = re.findall(pattern, str(parsed_expression))

    column_names = set(schema.field(int(field_path.split()[0])).name for field_path in field_paths)
    return column_names


def _generate_row_group_file_name(
    filename: str,
    row_group: int,
    row_indexes_ranges: list[list[int]],
    bbox: tuple[float, float, float, float],
) -> Path:
    import hashlib

    row_indexes_ranges_str = str(
        [
            (int(range_pair[0]), int(range_pair[1]))  # noqa: FURB123
            for range_pair in sorted(row_indexes_ranges, key=operator.itemgetter(0))
        ]
    )
    h = hashlib.new("sha256")
    h.update(row_indexes_ranges_str.encode())
    row_indexes_ranges_hash_part = h.hexdigest()

    h = hashlib.new("sha256")
    h.update(",".join(str(round(value, 7)) for value in bbox).encode())
    bbox_hash_part = h.hexdigest()

    stripped_filename_part = Path(filename.split("/", 2)[-1])
    stem = stripped_filename_part.stem.split(".")[0]
    stem = f"{stem}_{row_group}_{row_indexes_ranges_hash_part[:8]}_{bbox_hash_part[:8]}"
    return stripped_filename_part.with_stem(stem)


def _generate_result_file_path(
    release: str,
    theme: str,
    type: str,
    geometry_filter: "BaseGeometry",
    pyarrow_filter: Optional["Expression"],
    columns_to_download: Optional[list[str]],
    # keep_all_tags: bool,
    # explode_tags: bool,
    # filter_osm_ids: list[str],
    # save_as_wkt: bool,
) -> Path:
    import hashlib

    clipping_geometry_hash_part = _generate_geometry_hash(geometry_filter)

    pyarrow_filter_hash_part = "nofilter"
    if pyarrow_filter is not None:
        h = hashlib.new("sha256")
        h.update(str(pyarrow_filter).encode())
        pyarrow_filter_hash_part = h.hexdigest()[:8]

    columns_hash_part = ""
    if columns_to_download is not None:
        h = hashlib.new("sha256")
        h.update(str(sorted(columns_to_download)).encode())
        columns_hash_part = f"_{h.hexdigest()[:8]}"

    return (
        Path(release)
        / f"theme={theme}"
        / f"type={type}"
        / f"{clipping_geometry_hash_part}_{pyarrow_filter_hash_part}{columns_hash_part}.parquet"
    )


def _generate_geometry_hash(geometry_filter: "BaseGeometry") -> str:
    import hashlib

    import shapely.wkt as wktlib

    oriented_geometry = _get_oriented_geometry_filter(geometry_filter)
    h = hashlib.new("sha256")
    h.update(wktlib.dumps(oriented_geometry).encode())
    clipping_geometry_hash_part = h.hexdigest()

    return clipping_geometry_hash_part[:8]


def _get_oriented_geometry_filter(geometry_filter: "BaseGeometry") -> "BaseGeometry":
    import itertools

    from shapely import LinearRing, Polygon
    from shapely.geometry.base import BaseGeometry, BaseMultipartGeometry

    geometry = geometry_filter

    if isinstance(geometry, LinearRing):
        # https://stackoverflow.com/a/73073112/7766101
        new_coords = []
        if geometry.is_ccw:
            perimeter = list(geometry.coords)
        else:
            perimeter = list(geometry.coords)[::-1]
        smallest_point = sorted(perimeter)[0]
        double_iteration = itertools.chain(perimeter[:-1], perimeter)
        for point in double_iteration:
            if point == smallest_point:
                new_coords.append((round(point[0], 7), round(point[1], 7)))
                while len(new_coords) < len(perimeter):
                    next_point = next(double_iteration)
                    new_coords.append((round(next_point[0], 7), round(next_point[1], 7)))
                break
        return LinearRing(new_coords)
    if isinstance(geometry, Polygon):
        oriented_exterior = _get_oriented_geometry_filter(geometry.exterior)
        oriented_interiors = [
            cast(BaseGeometry, _get_oriented_geometry_filter(interior))
            for interior in geometry.interiors
        ]
        return Polygon(
            oriented_exterior,
            sorted(oriented_interiors, key=lambda geom: (geom.centroid.x, geom.centroid.y)),
        )
    elif isinstance(geometry, BaseMultipartGeometry):
        oriented_geoms = [
            cast(BaseGeometry, _get_oriented_geometry_filter(geom)) for geom in geometry.geoms
        ]
        return geometry.__class__(
            sorted(oriented_geoms, key=lambda geom: (geom.centroid.x, geom.centroid.y))
        )

    return geometry


def _filter_data_properly(
    parquet_filename: str,
    parquet_row_group: int,
    pyarrow_table: Any,
    geometry_filter: "BaseGeometry",
) -> Any:
    import geopandas as gpd
    from geoarrow.rust.core import GeometryArray
    from shapely import STRtree

    matching_indexes = STRtree(
        gpd.GeoSeries.from_arrow(
            GeometryArray.from_arrow(pyarrow_table["geometry"].combine_chunks())
        )
    ).query(geometry_filter, predicate="intersects")
    return pyarrow_table.take(matching_indexes)



"""Helper functions for GeoDataFrame IO operations."""

from pathlib import Path

import geopandas as gpd
import pyarrow.parquet as pq


def geodataframe_from_parquet(parquet_path: Path) -> gpd.GeoDataFrame:
    """
    Read parquet file as GeoDataframe.

    Loads the file without partitioning columns. Creates an index based on the id column.
    """
    schema = pq.ParquetDataset(parquet_path, partitioning=None).schema
    gdf = gpd.read_parquet(parquet_path, partitioning=None, columns=schema.names)
    if "id" in gdf.columns:
        gdf.set_index("id", inplace=True)

    return gdf



"""Functions for global cache."""

import shutil
from pathlib import Path

import platformdirs


def clear_release_indexes_cache() -> None:
    """Clear library release indexes cache."""
    for directory in (
        get_global_release_cache_directory(),
        get_global_wide_form_release_cache_directory(),
    ):
        shutil.rmtree(directory)


def get_global_release_cache_directory() -> Path:
    """Get global index cache location path."""
    return Path(platformdirs.user_cache_dir("OvertureMaestro")) / "release_indexes"


def _get_global_release_cache_directory(release: str) -> Path:
    return get_global_release_cache_directory() / release


def _get_local_release_cache_directory(release: str) -> Path:
    return Path("release_indexes") / release


def get_global_wide_form_release_cache_directory() -> Path:
    """Get global index cache location path."""
    return Path(platformdirs.user_cache_dir("OvertureMaestro")) / "wide_form_release_indexes"


def _get_global_wide_form_release_cache_directory(release: str) -> Path:
    return get_global_wide_form_release_cache_directory() / release


def _get_local_wide_form_release_cache_directory(release: str) -> Path:
    return Path("wide_form_release_indexes") / release



"""
Advanced Functions.

This module contains helper functions to simplify the API.
"""

from pathlib import Path
from typing import Optional, Union, overload

import geopandas as gpd
from shapely import box
from shapely.geometry.base import BaseGeometry

from overturemaestro._rich_progress import VERBOSITY_MODE
from overturemaestro.advanced_functions.wide_form import (
    convert_geometry_to_wide_form_parquet_for_all_types,
    convert_geometry_to_wide_form_parquet_for_multiple_types,
)
from overturemaestro.data_downloader import PYARROW_FILTER
from overturemaestro.geopandas_io import geodataframe_from_parquet

# TODO: add examples

__all__ = [
    "convert_bounding_box_to_wide_form_geodataframe",
    "convert_bounding_box_to_wide_form_geodataframe_for_all_types",
    "convert_bounding_box_to_wide_form_geodataframe_for_multiple_types",
    "convert_bounding_box_to_wide_form_parquet",
    "convert_bounding_box_to_wide_form_parquet_for_all_types",
    "convert_bounding_box_to_wide_form_parquet_for_multiple_types",
    "convert_geometry_to_wide_form_geodataframe",
    "convert_geometry_to_wide_form_geodataframe_for_all_types",
    "convert_geometry_to_wide_form_geodataframe_for_multiple_types",
    "convert_geometry_to_wide_form_parquet",
    "convert_geometry_to_wide_form_parquet_for_all_types",
    "convert_geometry_to_wide_form_parquet_for_multiple_types",
]


@overload
def convert_geometry_to_wide_form_parquet(
    theme: str,
    type: str,
    geometry_filter: BaseGeometry,
    *,
    include_all_possible_columns: bool = True,
    hierarchy_depth: Optional[int] = None,
    pyarrow_filters: Optional[PYARROW_FILTER] = None,
    result_file_path: Optional[Union[str, Path]] = None,
    ignore_cache: bool = False,
    working_directory: Union[str, Path] = "files",
    verbosity_mode: VERBOSITY_MODE = "transient",
    max_workers: Optional[int] = None,
    places_use_primary_category_only: bool = False,
    places_minimal_confidence: float = 0.75,
) -> Path: ...


@overload
def convert_geometry_to_wide_form_parquet(
    theme: str,
    type: str,
    geometry_filter: BaseGeometry,
    release: str,
    *,
    include_all_possible_columns: bool = True,
    hierarchy_depth: Optional[int] = None,
    pyarrow_filters: Optional[PYARROW_FILTER] = None,
    result_file_path: Optional[Union[str, Path]] = None,
    ignore_cache: bool = False,
    working_directory: Union[str, Path] = "files",
    verbosity_mode: VERBOSITY_MODE = "transient",
    max_workers: Optional[int] = None,
    places_use_primary_category_only: bool = False,
    places_minimal_confidence: float = 0.75,
) -> Path: ...


@overload
def convert_geometry_to_wide_form_parquet(
    theme: str,
    type: str,
    geometry_filter: BaseGeometry,
    release: Optional[str] = None,
    *,
    include_all_possible_columns: bool = True,
    hierarchy_depth: Optional[int] = None,
    pyarrow_filters: Optional[PYARROW_FILTER] = None,
    result_file_path: Optional[Union[str, Path]] = None,
    ignore_cache: bool = False,
    working_directory: Union[str, Path] = "files",
    verbosity_mode: VERBOSITY_MODE = "transient",
    max_workers: Optional[int] = None,
    places_use_primary_category_only: bool = False,
    places_minimal_confidence: float = 0.75,
) -> Path: ...


def convert_geometry_to_wide_form_parquet(
    theme: str,
    type: str,
    geometry_filter: BaseGeometry,
    release: Optional[str] = None,
    *,
    include_all_possible_columns: bool = True,
    hierarchy_depth: Optional[int] = None,
    pyarrow_filters: Optional[PYARROW_FILTER] = None,
    result_file_path: Optional[Union[str, Path]] = None,
    ignore_cache: bool = False,
    working_directory: Union[str, Path] = "files",
    verbosity_mode: VERBOSITY_MODE = "transient",
    max_workers: Optional[int] = None,
    places_use_primary_category_only: bool = False,
    places_minimal_confidence: float = 0.75,
) -> Path:
    """
    Get GeoParquet file for a given geometry in a wide format.

    Automatically downloads Overture Maps dataset for a given release and theme/type
    in a concurrent manner and returns a single file as a result with multiple columns based
    on dataset schema.

    Args:
        theme (str): Theme of the dataset.
        type (str): Type of the dataset.
        geometry_filter (BaseGeometry): Geometry used to filter data.
        release (Optional[str], optional): Release version. If not provided, will automatically load
            newest available release version. Defaults to None.
        include_all_possible_columns (bool, optional): Whether to have always the same list of
            columns in the resulting file. This ensures that always the same set of columns is
            returned for a given release for different regions. This also means, that some columns
            might be all filled with a False value. Defaults to True.
        hierarchy_depth (Optional[int], optional): Depth used to calculate how many hierarchy
            columns should be used to generate the wide form of the data. If None, will use all
            available columns. Defaults to None.
        pyarrow_filters (Optional[PYARROW_FILTER], optional): A pyarrow expression used to filter
            specific theme type pair. Defaults to None.
        result_file_path (Union[str, Path], optional): Where to save
            the geoparquet file. If not provided, will be generated based on hashes
            from filters. Defaults to None.
        ignore_cache (bool, optional): Whether to ignore precalculated geoparquet files or not.
            Defaults to False.
        working_directory (Union[str, Path], optional): Directory where to save
            the downloaded `*.parquet` files. Defaults to "files".
        verbosity_mode (Literal["silent", "transient", "verbose"], optional): Set progress
            verbosity mode. Can be one of: silent, transient and verbose. Silent disables
            output completely. Transient tracks progress, but removes output after finished.
            Verbose leaves all progress outputs in the stdout. Defaults to "transient".
        max_workers (Optional[int], optional): Max number of multiprocessing workers used to
            process the dataset. Defaults to None.
        places_use_primary_category_only (bool, optional): Whether to use only the primary category
            for places. Defaults to False.
        places_minimal_confidence (float, optional): Minimal confidence level for the places
            dataset. Defaults to 0.75.

    Returns:
        Path: Path to the generated GeoParquet file.
    """
    return convert_geometry_to_wide_form_parquet_for_multiple_types(
        theme_type_pairs=[(theme, type)],
        geometry_filter=geometry_filter,
        release=release,
        include_all_possible_columns=include_all_possible_columns,
        hierarchy_depth=hierarchy_depth,
        pyarrow_filters=[pyarrow_filters],
        result_file_path=result_file_path,
        ignore_cache=ignore_cache,
        working_directory=working_directory,
        verbosity_mode=verbosity_mode,
        max_workers=max_workers,
        places_use_primary_category_only=places_use_primary_category_only,
        places_minimal_confidence=places_minimal_confidence,
    )


@overload
def convert_geometry_to_wide_form_geodataframe(
    theme: str,
    type: str,
    geometry_filter: BaseGeometry,
    *,
    include_all_possible_columns: bool = True,
    hierarchy_depth: Optional[int] = None,
    pyarrow_filters: Optional[PYARROW_FILTER] = None,
    ignore_cache: bool = False,
    working_directory: Union[str, Path] = "files",
    verbosity_mode: VERBOSITY_MODE = "transient",
    max_workers: Optional[int] = None,
    places_use_primary_category_only: bool = False,
    places_minimal_confidence: float = 0.75,
) -> gpd.GeoDataFrame: ...


@overload
def convert_geometry_to_wide_form_geodataframe(
    theme: str,
    type: str,
    geometry_filter: BaseGeometry,
    release: str,
    *,
    include_all_possible_columns: bool = True,
    hierarchy_depth: Optional[int] = None,
    pyarrow_filters: Optional[PYARROW_FILTER] = None,
    ignore_cache: bool = False,
    working_directory: Union[str, Path] = "files",
    verbosity_mode: VERBOSITY_MODE = "transient",
    max_workers: Optional[int] = None,
    places_use_primary_category_only: bool = False,
    places_minimal_confidence: float = 0.75,
) -> gpd.GeoDataFrame: ...


@overload
def convert_geometry_to_wide_form_geodataframe(
    theme: str,
    type: str,
    geometry_filter: BaseGeometry,
    release: Optional[str] = None,
    *,
    include_all_possible_columns: bool = True,
    hierarchy_depth: Optional[int] = None,
    pyarrow_filters: Optional[PYARROW_FILTER] = None,
    ignore_cache: bool = False,
    working_directory: Union[str, Path] = "files",
    verbosity_mode: VERBOSITY_MODE = "transient",
    max_workers: Optional[int] = None,
    places_use_primary_category_only: bool = False,
    places_minimal_confidence: float = 0.75,
) -> gpd.GeoDataFrame: ...


def convert_geometry_to_wide_form_geodataframe(
    theme: str,
    type: str,
    geometry_filter: BaseGeometry,
    release: Optional[str] = None,
    *,
    include_all_possible_columns: bool = True,
    hierarchy_depth: Optional[int] = None,
    pyarrow_filters: Optional[PYARROW_FILTER] = None,
    ignore_cache: bool = False,
    working_directory: Union[str, Path] = "files",
    verbosity_mode: VERBOSITY_MODE = "transient",
    max_workers: Optional[int] = None,
    places_use_primary_category_only: bool = False,
    places_minimal_confidence: float = 0.75,
) -> gpd.GeoDataFrame:
    """
    Get GeoDataFrame for a given geometry in a wide format.

    Automatically downloads Overture Maps dataset for a given release and theme/type
    in a concurrent manner and returns a single GeoDataFrame as a result with multiple columns based
    on dataset schema.

    Args:
        theme (str): Theme of the dataset.
        type (str): Type of the dataset.
        geometry_filter (BaseGeometry): Geometry used to filter data.
        release (Optional[str], optional): Release version. If not provided, will automatically load
            newest available release version. Defaults to None.
        include_all_possible_columns (bool, optional): Whether to have always the same list of
            columns in the resulting file. This ensures that always the same set of columns is
            returned for a given release for different regions. This also means, that some columns
            might be all filled with a False value. Defaults to True.
        hierarchy_depth (Optional[int], optional): Depth used to calculate how many hierarchy
            columns should be used to generate the wide form of the data. If None, will use all
            available columns. Defaults to None.
        pyarrow_filters (Optional[PYARROW_FILTER], optional): A pyarrow expression used to filter
            specific theme type pair. Defaults to None.
        ignore_cache (bool, optional): Whether to ignore precalculated geoparquet files or not.
            Defaults to False.
        working_directory (Union[str, Path], optional): Directory where to save
            the downloaded `*.parquet` files. Defaults to "files".
        verbosity_mode (Literal["silent", "transient", "verbose"], optional): Set progress
            verbosity mode. Can be one of: silent, transient and verbose. Silent disables
            output completely. Transient tracks progress, but removes output after finished.
            Verbose leaves all progress outputs in the stdout. Defaults to "transient".
        max_workers (Optional[int], optional): Max number of multiprocessing workers used to
            process the dataset. Defaults to None.
        places_use_primary_category_only (bool, optional): Whether to use only the primary category
            for places. Defaults to False.
        places_minimal_confidence (float, optional): Minimal confidence level for the places
            dataset. Defaults to 0.75.

    Returns:
        gpd.GeoDataFrame: GeoDataFrame with Overture Maps features.
    """
    parsed_geoparquet_file = convert_geometry_to_wide_form_parquet(
        theme=theme,
        type=type,
        geometry_filter=geometry_filter,
        release=release,
        include_all_possible_columns=include_all_possible_columns,
        hierarchy_depth=hierarchy_depth,
        pyarrow_filters=pyarrow_filters,
        ignore_cache=ignore_cache,
        working_directory=working_directory,
        verbosity_mode=verbosity_mode,
        max_workers=max_workers,
        places_use_primary_category_only=places_use_primary_category_only,
        places_minimal_confidence=places_minimal_confidence,
    )
    return geodataframe_from_parquet(parsed_geoparquet_file)


@overload
def convert_geometry_to_wide_form_geodataframe_for_all_types(
    geometry_filter: BaseGeometry,
    *,
    include_all_possible_columns: bool = True,
    hierarchy_depth: Optional[Union[int, list[Optional[int]]]] = None,
    pyarrow_filters: Optional[list[Optional[PYARROW_FILTER]]] = None,
    ignore_cache: bool = False,
    working_directory: Union[str, Path] = "files",
    verbosity_mode: VERBOSITY_MODE = "transient",
    max_workers: Optional[int] = None,
    places_use_primary_category_only: bool = False,
    places_minimal_confidence: float = 0.75,
) -> gpd.GeoDataFrame: ...


@overload
def convert_geometry_to_wide_form_geodataframe_for_all_types(
    geometry_filter: BaseGeometry,
    release: str,
    *,
    include_all_possible_columns: bool = True,
    hierarchy_depth: Optional[Union[int, list[Optional[int]]]] = None,
    pyarrow_filters: Optional[list[Optional[PYARROW_FILTER]]] = None,
    ignore_cache: bool = False,
    working_directory: Union[str, Path] = "files",
    verbosity_mode: VERBOSITY_MODE = "transient",
    max_workers: Optional[int] = None,
    places_use_primary_category_only: bool = False,
    places_minimal_confidence: float = 0.75,
) -> gpd.GeoDataFrame: ...


@overload
def convert_geometry_to_wide_form_geodataframe_for_all_types(
    geometry_filter: BaseGeometry,
    release: Optional[str] = None,
    *,
    include_all_possible_columns: bool = True,
    hierarchy_depth: Optional[Union[int, list[Optional[int]]]] = None,
    pyarrow_filters: Optional[list[Optional[PYARROW_FILTER]]] = None,
    ignore_cache: bool = False,
    working_directory: Union[str, Path] = "files",
    verbosity_mode: VERBOSITY_MODE = "transient",
    max_workers: Optional[int] = None,
    places_use_primary_category_only: bool = False,
    places_minimal_confidence: float = 0.75,
) -> gpd.GeoDataFrame: ...


def convert_geometry_to_wide_form_geodataframe_for_all_types(
    geometry_filter: BaseGeometry,
    release: Optional[str] = None,
    *,
    include_all_possible_columns: bool = True,
    hierarchy_depth: Optional[Union[int, list[Optional[int]]]] = None,
    pyarrow_filters: Optional[list[Optional[PYARROW_FILTER]]] = None,
    ignore_cache: bool = False,
    working_directory: Union[str, Path] = "files",
    verbosity_mode: VERBOSITY_MODE = "transient",
    max_workers: Optional[int] = None,
    places_use_primary_category_only: bool = False,
    places_minimal_confidence: float = 0.75,
) -> gpd.GeoDataFrame:
    """
    Get GeoDataFrame for a given geometry in a wide format for all types.

    Automatically downloads Overture Maps dataset for a given release and all available theme/types
    in a concurrent manner and returns a single GeoDataFrame as a result with multiple columns based
    on dataset schema.

    Args:
        geometry_filter (BaseGeometry): Geometry used to filter data.
        release (Optional[str], optional): Release version. If not provided, will automatically load
            newest available release version. Defaults to None.
        include_all_possible_columns (bool, optional): Whether to have always the same list of
            columns in the resulting file. This ensures that always the same set of columns is
            returned for a given release for different regions. This also means, that some columns
            might be all filled with a False value. Defaults to True.
        hierarchy_depth (Optional[Union[int, list[Optional[int]]]], optional): Depth used to
            calculate how many hierarchy columns should be used to generate the wide form of
            the data. Can be a single integer or a list of integers. If None, will use all
            available columns. Defaults to None.
        pyarrow_filters (Optional[list[Optional[PYARROW_FILTER]]], optional): A list of pyarrow
            expressions used to filter specific theme type pair. Must be the same length as the list
            of theme type pairs. Defaults to None.
        ignore_cache (bool, optional): Whether to ignore precalculated geoparquet files or not.
            Defaults to False.
        working_directory (Union[str, Path], optional): Directory where to save
            the downloaded `*.parquet` files. Defaults to "files".
        verbosity_mode (Literal["silent", "transient", "verbose"], optional): Set progress
            verbosity mode. Can be one of: silent, transient and verbose. Silent disables
            output completely. Transient tracks progress, but removes output after finished.
            Verbose leaves all progress outputs in the stdout. Defaults to "transient".
        max_workers (Optional[int], optional): Max number of multiprocessing workers used to
            process the dataset. Defaults to None.
        places_use_primary_category_only (bool, optional): Whether to use only the primary category
            for places. Defaults to False.
        places_minimal_confidence (float, optional): Minimal confidence level for the places
            dataset. Defaults to 0.75.

    Returns:
        gpd.GeoDataFrame: GeoDataFrame with Overture Maps features.
    """
    parsed_geoparquet_file = convert_geometry_to_wide_form_parquet_for_all_types(
        geometry_filter=geometry_filter,
        release=release,
        include_all_possible_columns=include_all_possible_columns,
        hierarchy_depth=hierarchy_depth,
        pyarrow_filters=pyarrow_filters,
        ignore_cache=ignore_cache,
        working_directory=working_directory,
        verbosity_mode=verbosity_mode,
        max_workers=max_workers,
        places_use_primary_category_only=places_use_primary_category_only,
        places_minimal_confidence=places_minimal_confidence,
    )
    return geodataframe_from_parquet(parsed_geoparquet_file)


@overload
def convert_geometry_to_wide_form_geodataframe_for_multiple_types(
    theme_type_pairs: list[tuple[str, str]],
    geometry_filter: BaseGeometry,
    *,
    include_all_possible_columns: bool = True,
    hierarchy_depth: Optional[Union[int, list[Optional[int]]]] = None,
    pyarrow_filters: Optional[list[Optional[PYARROW_FILTER]]] = None,
    ignore_cache: bool = False,
    working_directory: Union[str, Path] = "files",
    verbosity_mode: VERBOSITY_MODE = "transient",
    max_workers: Optional[int] = None,
    places_use_primary_category_only: bool = False,
    places_minimal_confidence: float = 0.75,
) -> gpd.GeoDataFrame: ...


@overload
def convert_geometry_to_wide_form_geodataframe_for_multiple_types(
    theme_type_pairs: list[tuple[str, str]],
    geometry_filter: BaseGeometry,
    release: str,
    *,
    include_all_possible_columns: bool = True,
    hierarchy_depth: Optional[Union[int, list[Optional[int]]]] = None,
    pyarrow_filters: Optional[list[Optional[PYARROW_FILTER]]] = None,
    ignore_cache: bool = False,
    working_directory: Union[str, Path] = "files",
    verbosity_mode: VERBOSITY_MODE = "transient",
    max_workers: Optional[int] = None,
    places_use_primary_category_only: bool = False,
    places_minimal_confidence: float = 0.75,
) -> gpd.GeoDataFrame: ...


@overload
def convert_geometry_to_wide_form_geodataframe_for_multiple_types(
    theme_type_pairs: list[tuple[str, str]],
    geometry_filter: BaseGeometry,
    release: Optional[str] = None,
    *,
    include_all_possible_columns: bool = True,
    hierarchy_depth: Optional[Union[int, list[Optional[int]]]] = None,
    pyarrow_filters: Optional[list[Optional[PYARROW_FILTER]]] = None,
    ignore_cache: bool = False,
    working_directory: Union[str, Path] = "files",
    verbosity_mode: VERBOSITY_MODE = "transient",
    max_workers: Optional[int] = None,
    places_use_primary_category_only: bool = False,
    places_minimal_confidence: float = 0.75,
) -> gpd.GeoDataFrame: ...


def convert_geometry_to_wide_form_geodataframe_for_multiple_types(
    theme_type_pairs: list[tuple[str, str]],
    geometry_filter: BaseGeometry,
    release: Optional[str] = None,
    *,
    include_all_possible_columns: bool = True,
    hierarchy_depth: Optional[Union[int, list[Optional[int]]]] = None,
    pyarrow_filters: Optional[list[Optional[PYARROW_FILTER]]] = None,
    ignore_cache: bool = False,
    working_directory: Union[str, Path] = "files",
    verbosity_mode: VERBOSITY_MODE = "transient",
    max_workers: Optional[int] = None,
    places_use_primary_category_only: bool = False,
    places_minimal_confidence: float = 0.75,
) -> gpd.GeoDataFrame:
    """
    Get GeoDataFrame for a given geometry in a wide format for multiple types.

    Automatically downloads Overture Maps dataset for a given release and theme/type pairs
    in a concurrent manner and returns a single GeoDataFrame as a result with multiple columns based
    on dataset schema.

    Args:
        theme_type_pairs (list[tuple[str, str]]): Pairs of themes and types of the dataset.
        geometry_filter (BaseGeometry): Geometry used to filter data.
        release (Optional[str], optional): Release version. If not provided, will automatically load
            newest available release version. Defaults to None.
        include_all_possible_columns (bool, optional): Whether to have always the same list of
            columns in the resulting file. This ensures that always the same set of columns is
            returned for a given release for different regions. This also means, that some columns
            might be all filled with a False value. Defaults to True.
        hierarchy_depth (Optional[Union[int, list[Optional[int]]]], optional): Depth used to
            calculate how many hierarchy columns should be used to generate the wide form of
            the data. Can be a single integer or a list of integers. If None, will use all
            available columns. Defaults to None.
        pyarrow_filters (Optional[list[Optional[PYARROW_FILTER]]], optional): A list of pyarrow
            expressions used to filter specific theme type pair. Must be the same length as the list
            of theme type pairs. Defaults to None.
        ignore_cache (bool, optional): Whether to ignore precalculated geoparquet files or not.
            Defaults to False.
        working_directory (Union[str, Path], optional): Directory where to save
            the downloaded `*.parquet` files. Defaults to "files".
        verbosity_mode (Literal["silent", "transient", "verbose"], optional): Set progress
            verbosity mode. Can be one of: silent, transient and verbose. Silent disables
            output completely. Transient tracks progress, but removes output after finished.
            Verbose leaves all progress outputs in the stdout. Defaults to "transient".
        max_workers (Optional[int], optional): Max number of multiprocessing workers used to
            process the dataset. Defaults to None.
        places_use_primary_category_only (bool, optional): Whether to use only the primary category
            for places. Defaults to False.
        places_minimal_confidence (float, optional): Minimal confidence level for the places
            dataset. Defaults to 0.75.

    Returns:
        gpd.GeoDataFrame: GeoDataFrame with Overture Maps features.
    """
    parsed_geoparquet_file = convert_geometry_to_wide_form_parquet_for_multiple_types(
        theme_type_pairs=theme_type_pairs,
        geometry_filter=geometry_filter,
        release=release,
        include_all_possible_columns=include_all_possible_columns,
        hierarchy_depth=hierarchy_depth,
        pyarrow_filters=pyarrow_filters,
        ignore_cache=ignore_cache,
        working_directory=working_directory,
        verbosity_mode=verbosity_mode,
        max_workers=max_workers,
        places_use_primary_category_only=places_use_primary_category_only,
        places_minimal_confidence=places_minimal_confidence,
    )
    return geodataframe_from_parquet(parsed_geoparquet_file)


@overload
def convert_bounding_box_to_wide_form_parquet(
    theme: str,
    type: str,
    bbox: tuple[float, float, float, float],
    *,
    include_all_possible_columns: bool = True,
    hierarchy_depth: Optional[int] = None,
    pyarrow_filters: Optional[PYARROW_FILTER] = None,
    result_file_path: Optional[Union[str, Path]] = None,
    ignore_cache: bool = False,
    working_directory: Union[str, Path] = "files",
    verbosity_mode: VERBOSITY_MODE = "transient",
    max_workers: Optional[int] = None,
    places_use_primary_category_only: bool = False,
    places_minimal_confidence: float = 0.75,
) -> Path: ...


@overload
def convert_bounding_box_to_wide_form_parquet(
    theme: str,
    type: str,
    bbox: tuple[float, float, float, float],
    release: str,
    *,
    include_all_possible_columns: bool = True,
    hierarchy_depth: Optional[int] = None,
    pyarrow_filters: Optional[PYARROW_FILTER] = None,
    result_file_path: Optional[Union[str, Path]] = None,
    ignore_cache: bool = False,
    working_directory: Union[str, Path] = "files",
    verbosity_mode: VERBOSITY_MODE = "transient",
    max_workers: Optional[int] = None,
    places_use_primary_category_only: bool = False,
    places_minimal_confidence: float = 0.75,
) -> Path: ...


@overload
def convert_bounding_box_to_wide_form_parquet(
    theme: str,
    type: str,
    bbox: tuple[float, float, float, float],
    release: Optional[str] = None,
    *,
    include_all_possible_columns: bool = True,
    hierarchy_depth: Optional[int] = None,
    pyarrow_filters: Optional[PYARROW_FILTER] = None,
    result_file_path: Optional[Union[str, Path]] = None,
    ignore_cache: bool = False,
    working_directory: Union[str, Path] = "files",
    verbosity_mode: VERBOSITY_MODE = "transient",
    max_workers: Optional[int] = None,
    places_use_primary_category_only: bool = False,
    places_minimal_confidence: float = 0.75,
) -> Path: ...


def convert_bounding_box_to_wide_form_parquet(
    theme: str,
    type: str,
    bbox: tuple[float, float, float, float],
    release: Optional[str] = None,
    *,
    include_all_possible_columns: bool = True,
    hierarchy_depth: Optional[int] = None,
    pyarrow_filters: Optional[PYARROW_FILTER] = None,
    result_file_path: Optional[Union[str, Path]] = None,
    ignore_cache: bool = False,
    working_directory: Union[str, Path] = "files",
    verbosity_mode: VERBOSITY_MODE = "transient",
    max_workers: Optional[int] = None,
    places_use_primary_category_only: bool = False,
    places_minimal_confidence: float = 0.75,
) -> Path:
    """
    Get GeoParquet file for a given bounding box in a wide format.

    Automatically downloads Overture Maps dataset for a given release and theme/type
    in a concurrent manner and returns a single file as a result with multiple columns based
    on dataset schema.

    Args:
        theme (str): Theme of the dataset.
        type (str): Type of the dataset.
        bbox (tuple[float, float, float, float]): Bounding box used to filter data.
            Order of values: xmin, ymin, xmax, ymax.
        release (Optional[str], optional): Release version. If not provided, will automatically load
            newest available release version. Defaults to None.
        include_all_possible_columns (bool, optional): Whether to have always the same list of
            columns in the resulting file. This ensures that always the same set of columns is
            returned for a given release for different regions. This also means, that some columns
            might be all filled with a False value. Defaults to True.
        hierarchy_depth (Optional[int], optional): Depth used to calculate how many hierarchy
            columns should be used to generate the wide form of the data. If None, will use all
            available columns. Defaults to None.
        pyarrow_filters (Optional[PYARROW_FILTER], optional): A pyarrow
            expression used to filter specific theme type pair. Defaults to None.
        result_file_path (Union[str, Path], optional): Where to save
            the geoparquet file. If not provided, will be generated based on hashes
            from filters. Defaults to None.
        ignore_cache (bool, optional): Whether to ignore precalculated geoparquet files or not.
            Defaults to False.
        working_directory (Union[str, Path], optional): Directory where to save
            the downloaded `*.parquet` files. Defaults to "files".
        verbosity_mode (Literal["silent", "transient", "verbose"], optional): Set progress
            verbosity mode. Can be one of: silent, transient and verbose. Silent disables
            output completely. Transient tracks progress, but removes output after finished.
            Verbose leaves all progress outputs in the stdout. Defaults to "transient".
        max_workers (Optional[int], optional): Max number of multiprocessing workers used to
            process the dataset. Defaults to None.
        places_use_primary_category_only (bool, optional): Whether to use only the primary category
            for places. Defaults to False.
        places_minimal_confidence (float, optional): Minimal confidence level for the places
            dataset. Defaults to 0.75.

    Returns:
        Path: Path to the generated GeoParquet file.
    """
    return convert_geometry_to_wide_form_parquet(
        theme=theme,
        type=type,
        geometry_filter=box(*bbox),
        release=release,
        include_all_possible_columns=include_all_possible_columns,
        hierarchy_depth=hierarchy_depth,
        pyarrow_filters=pyarrow_filters,
        result_file_path=result_file_path,
        ignore_cache=ignore_cache,
        working_directory=working_directory,
        verbosity_mode=verbosity_mode,
        max_workers=max_workers,
        places_use_primary_category_only=places_use_primary_category_only,
        places_minimal_confidence=places_minimal_confidence,
    )


@overload
def convert_bounding_box_to_wide_form_parquet_for_all_types(
    bbox: tuple[float, float, float, float],
    *,
    include_all_possible_columns: bool = True,
    hierarchy_depth: Optional[Union[int, list[Optional[int]]]] = None,
    pyarrow_filters: Optional[list[Optional[PYARROW_FILTER]]] = None,
    result_file_path: Optional[Union[str, Path]] = None,
    ignore_cache: bool = False,
    working_directory: Union[str, Path] = "files",
    verbosity_mode: VERBOSITY_MODE = "transient",
    max_workers: Optional[int] = None,
    places_use_primary_category_only: bool = False,
    places_minimal_confidence: float = 0.75,
) -> Path: ...


@overload
def convert_bounding_box_to_wide_form_parquet_for_all_types(
    bbox: tuple[float, float, float, float],
    release: str,
    *,
    include_all_possible_columns: bool = True,
    hierarchy_depth: Optional[Union[int, list[Optional[int]]]] = None,
    pyarrow_filters: Optional[list[Optional[PYARROW_FILTER]]] = None,
    result_file_path: Optional[Union[str, Path]] = None,
    ignore_cache: bool = False,
    working_directory: Union[str, Path] = "files",
    verbosity_mode: VERBOSITY_MODE = "transient",
    max_workers: Optional[int] = None,
    places_use_primary_category_only: bool = False,
    places_minimal_confidence: float = 0.75,
) -> Path: ...


@overload
def convert_bounding_box_to_wide_form_parquet_for_all_types(
    bbox: tuple[float, float, float, float],
    release: Optional[str] = None,
    *,
    include_all_possible_columns: bool = True,
    hierarchy_depth: Optional[Union[int, list[Optional[int]]]] = None,
    pyarrow_filters: Optional[list[Optional[PYARROW_FILTER]]] = None,
    result_file_path: Optional[Union[str, Path]] = None,
    ignore_cache: bool = False,
    working_directory: Union[str, Path] = "files",
    verbosity_mode: VERBOSITY_MODE = "transient",
    max_workers: Optional[int] = None,
    places_use_primary_category_only: bool = False,
    places_minimal_confidence: float = 0.75,
) -> Path: ...


def convert_bounding_box_to_wide_form_parquet_for_all_types(
    bbox: tuple[float, float, float, float],
    release: Optional[str] = None,
    *,
    include_all_possible_columns: bool = True,
    hierarchy_depth: Optional[Union[int, list[Optional[int]]]] = None,
    pyarrow_filters: Optional[list[Optional[PYARROW_FILTER]]] = None,
    result_file_path: Optional[Union[str, Path]] = None,
    ignore_cache: bool = False,
    working_directory: Union[str, Path] = "files",
    verbosity_mode: VERBOSITY_MODE = "transient",
    max_workers: Optional[int] = None,
    places_use_primary_category_only: bool = False,
    places_minimal_confidence: float = 0.75,
) -> Path:
    """
    Get GeoParquet file for a given bounding box in a wide format for all types.

    Automatically downloads Overture Maps dataset for a given release and all available theme/types
    in a concurrent manner and returns a single file as a result with multiple columns based
    on dataset schema.

    Args:
        bbox (tuple[float, float, float, float]): Bounding box used to filter data.
            Order of values: xmin, ymin, xmax, ymax.
        release (Optional[str], optional): Release version. If not provided, will automatically load
            newest available release version. Defaults to None.
        include_all_possible_columns (bool, optional): Whether to have always the same list of
            columns in the resulting file. This ensures that always the same set of columns is
            returned for a given release for different regions. This also means, that some columns
            might be all filled with a False value. Defaults to True.
        hierarchy_depth (Optional[Union[int, list[Optional[int]]]], optional): Depth used to
            calculate how many hierarchy columns should be used to generate the wide form of
            the data. Can be a single integer or a list of integers. If None, will use all
            available columns. Defaults to None.
        pyarrow_filters (Optional[list[Optional[PYARROW_FILTER]]], optional): A list of pyarrow
            expressions used to filter specific theme type pair. Must be the same length as the list
            of theme type pairs. Defaults to None.
        result_file_path (Union[str, Path], optional): Where to save
            the geoparquet file. If not provided, will be generated based on hashes
            from filters. Defaults to None.
        ignore_cache (bool, optional): Whether to ignore precalculated geoparquet files or not.
            Defaults to False.
        working_directory (Union[str, Path], optional): Directory where to save
            the downloaded `*.parquet` files. Defaults to "files".
        verbosity_mode (Literal["silent", "transient", "verbose"], optional): Set progress
            verbosity mode. Can be one of: silent, transient and verbose. Silent disables
            output completely. Transient tracks progress, but removes output after finished.
            Verbose leaves all progress outputs in the stdout. Defaults to "transient".
        max_workers (Optional[int], optional): Max number of multiprocessing workers used to
            process the dataset. Defaults to None.
        places_use_primary_category_only (bool, optional): Whether to use only the primary category
            for places. Defaults to False.
        places_minimal_confidence (float, optional): Minimal confidence level for the places
            dataset. Defaults to 0.75.

    Returns:
        Path: Path to the generated GeoParquet file.
    """
    return convert_geometry_to_wide_form_parquet_for_all_types(
        geometry_filter=box(*bbox),
        release=release,
        include_all_possible_columns=include_all_possible_columns,
        hierarchy_depth=hierarchy_depth,
        pyarrow_filters=pyarrow_filters,
        result_file_path=result_file_path,
        ignore_cache=ignore_cache,
        working_directory=working_directory,
        verbosity_mode=verbosity_mode,
        max_workers=max_workers,
        places_use_primary_category_only=places_use_primary_category_only,
        places_minimal_confidence=places_minimal_confidence,
    )


@overload
def convert_bounding_box_to_wide_form_parquet_for_multiple_types(
    theme_type_pairs: list[tuple[str, str]],
    bbox: tuple[float, float, float, float],
    *,
    include_all_possible_columns: bool = True,
    hierarchy_depth: Optional[Union[int, list[Optional[int]]]] = None,
    pyarrow_filters: Optional[list[Optional[PYARROW_FILTER]]] = None,
    result_file_path: Optional[Union[str, Path]] = None,
    ignore_cache: bool = False,
    working_directory: Union[str, Path] = "files",
    verbosity_mode: VERBOSITY_MODE = "transient",
    max_workers: Optional[int] = None,
    places_use_primary_category_only: bool = False,
    places_minimal_confidence: float = 0.75,
) -> Path: ...


@overload
def convert_bounding_box_to_wide_form_parquet_for_multiple_types(
    theme_type_pairs: list[tuple[str, str]],
    bbox: tuple[float, float, float, float],
    release: str,
    *,
    include_all_possible_columns: bool = True,
    hierarchy_depth: Optional[Union[int, list[Optional[int]]]] = None,
    pyarrow_filters: Optional[list[Optional[PYARROW_FILTER]]] = None,
    result_file_path: Optional[Union[str, Path]] = None,
    ignore_cache: bool = False,
    working_directory: Union[str, Path] = "files",
    verbosity_mode: VERBOSITY_MODE = "transient",
    max_workers: Optional[int] = None,
    places_use_primary_category_only: bool = False,
    places_minimal_confidence: float = 0.75,
) -> Path: ...


@overload
def convert_bounding_box_to_wide_form_parquet_for_multiple_types(
    theme_type_pairs: list[tuple[str, str]],
    bbox: tuple[float, float, float, float],
    release: Optional[str] = None,
    *,
    include_all_possible_columns: bool = True,
    hierarchy_depth: Optional[Union[int, list[Optional[int]]]] = None,
    pyarrow_filters: Optional[list[Optional[PYARROW_FILTER]]] = None,
    result_file_path: Optional[Union[str, Path]] = None,
    ignore_cache: bool = False,
    working_directory: Union[str, Path] = "files",
    verbosity_mode: VERBOSITY_MODE = "transient",
    max_workers: Optional[int] = None,
    places_use_primary_category_only: bool = False,
    places_minimal_confidence: float = 0.75,
) -> Path: ...


def convert_bounding_box_to_wide_form_parquet_for_multiple_types(
    theme_type_pairs: list[tuple[str, str]],
    bbox: tuple[float, float, float, float],
    release: Optional[str] = None,
    *,
    include_all_possible_columns: bool = True,
    hierarchy_depth: Optional[Union[int, list[Optional[int]]]] = None,
    pyarrow_filters: Optional[list[Optional[PYARROW_FILTER]]] = None,
    result_file_path: Optional[Union[str, Path]] = None,
    ignore_cache: bool = False,
    working_directory: Union[str, Path] = "files",
    verbosity_mode: VERBOSITY_MODE = "transient",
    max_workers: Optional[int] = None,
    places_use_primary_category_only: bool = False,
    places_minimal_confidence: float = 0.75,
) -> Path:
    """
    Get GeoParquet file for a given bounding box in a wide format for multiple types.

    Automatically downloads Overture Maps dataset for a given release and theme/type pairs
    in a concurrent manner and returns a single file as a result with multiple columns based
    on dataset schema.

    Args:
        theme_type_pairs (list[tuple[str, str]]): Pairs of themes and types of the dataset.
        bbox (tuple[float, float, float, float]): Bounding box used to filter data.
            Order of values: xmin, ymin, xmax, ymax.
        release (Optional[str], optional): Release version. If not provided, will automatically load
            newest available release version. Defaults to None.
        include_all_possible_columns (bool, optional): Whether to have always the same list of
            columns in the resulting file. This ensures that always the same set of columns is
            returned for a given release for different regions. This also means, that some columns
            might be all filled with a False value. Defaults to True.
        hierarchy_depth (Optional[Union[int, list[Optional[int]]]], optional): Depth used to
            calculate how many hierarchy columns should be used to generate the wide form of
            the data. Can be a single integer or a list of integers. If None, will use all
            available columns. Defaults to None.
        pyarrow_filters (Optional[list[Optional[PYARROW_FILTER]]], optional): A list of pyarrow
            expressions used to filter specific theme type pair. Must be the same length as the list
            of theme type pairs. Defaults to None.
        result_file_path (Union[str, Path], optional): Where to save
            the geoparquet file. If not provided, will be generated based on hashes
            from filters. Defaults to None.
        ignore_cache (bool, optional): Whether to ignore precalculated geoparquet files or not.
            Defaults to False.
        working_directory (Union[str, Path], optional): Directory where to save
            the downloaded `*.parquet` files. Defaults to "files".
        verbosity_mode (Literal["silent", "transient", "verbose"], optional): Set progress
            verbosity mode. Can be one of: silent, transient and verbose. Silent disables
            output completely. Transient tracks progress, but removes output after finished.
            Verbose leaves all progress outputs in the stdout. Defaults to "transient".
        max_workers (Optional[int], optional): Max number of multiprocessing workers used to
            process the dataset. Defaults to None.
        places_use_primary_category_only (bool, optional): Whether to use only the primary category
            for places. Defaults to False.
        places_minimal_confidence (float, optional): Minimal confidence level for the places
            dataset. Defaults to 0.75.

    Returns:
        Path: Path to the generated GeoParquet file.
    """
    return convert_geometry_to_wide_form_parquet_for_multiple_types(
        theme_type_pairs=theme_type_pairs,
        geometry_filter=box(*bbox),
        release=release,
        include_all_possible_columns=include_all_possible_columns,
        hierarchy_depth=hierarchy_depth,
        pyarrow_filters=pyarrow_filters,
        result_file_path=result_file_path,
        ignore_cache=ignore_cache,
        working_directory=working_directory,
        verbosity_mode=verbosity_mode,
        max_workers=max_workers,
        places_use_primary_category_only=places_use_primary_category_only,
        places_minimal_confidence=places_minimal_confidence,
    )


@overload
def convert_bounding_box_to_wide_form_geodataframe(
    theme: str,
    type: str,
    bbox: tuple[float, float, float, float],
    *,
    include_all_possible_columns: bool = True,
    hierarchy_depth: Optional[int] = None,
    pyarrow_filters: Optional[PYARROW_FILTER] = None,
    ignore_cache: bool = False,
    working_directory: Union[str, Path] = "files",
    verbosity_mode: VERBOSITY_MODE = "transient",
    max_workers: Optional[int] = None,
    places_use_primary_category_only: bool = False,
    places_minimal_confidence: float = 0.75,
) -> gpd.GeoDataFrame: ...


@overload
def convert_bounding_box_to_wide_form_geodataframe(
    theme: str,
    type: str,
    bbox: tuple[float, float, float, float],
    release: str,
    *,
    include_all_possible_columns: bool = True,
    hierarchy_depth: Optional[int] = None,
    pyarrow_filters: Optional[PYARROW_FILTER] = None,
    ignore_cache: bool = False,
    working_directory: Union[str, Path] = "files",
    verbosity_mode: VERBOSITY_MODE = "transient",
    max_workers: Optional[int] = None,
    places_use_primary_category_only: bool = False,
    places_minimal_confidence: float = 0.75,
) -> gpd.GeoDataFrame: ...


@overload
def convert_bounding_box_to_wide_form_geodataframe(
    theme: str,
    type: str,
    bbox: tuple[float, float, float, float],
    release: Optional[str] = None,
    *,
    include_all_possible_columns: bool = True,
    hierarchy_depth: Optional[int] = None,
    pyarrow_filters: Optional[PYARROW_FILTER] = None,
    ignore_cache: bool = False,
    working_directory: Union[str, Path] = "files",
    verbosity_mode: VERBOSITY_MODE = "transient",
    max_workers: Optional[int] = None,
    places_use_primary_category_only: bool = False,
    places_minimal_confidence: float = 0.75,
) -> gpd.GeoDataFrame: ...


def convert_bounding_box_to_wide_form_geodataframe(
    theme: str,
    type: str,
    bbox: tuple[float, float, float, float],
    release: Optional[str] = None,
    *,
    include_all_possible_columns: bool = True,
    hierarchy_depth: Optional[int] = None,
    pyarrow_filters: Optional[PYARROW_FILTER] = None,
    ignore_cache: bool = False,
    working_directory: Union[str, Path] = "files",
    verbosity_mode: VERBOSITY_MODE = "transient",
    max_workers: Optional[int] = None,
    places_use_primary_category_only: bool = False,
    places_minimal_confidence: float = 0.75,
) -> gpd.GeoDataFrame:
    """
    Get GeoDataFrame for a given bounding box in a wide format.

    Automatically downloads Overture Maps dataset for a given release and theme/type
    in a concurrent manner and returns a single GeoDataFrame as a result with multiple columns based
    on dataset schema.

    Args:
        theme (str): Theme of the dataset.
        type (str): Type of the dataset.
        bbox (tuple[float, float, float, float]): Bounding box used to filter data.
            Order of values: xmin, ymin, xmax, ymax.
        release (Optional[str], optional): Release version. If not provided, will automatically load
            newest available release version. Defaults to None.
        include_all_possible_columns (bool, optional): Whether to have always the same list of
            columns in the resulting file. This ensures that always the same set of columns is
            returned for a given release for different regions. This also means, that some columns
            might be all filled with a False value. Defaults to True.
        hierarchy_depth (Optional[int], optional): Depth used to calculate how many hierarchy
            columns should be used to generate the wide form of the data. If None, will use all
            available columns. Defaults to None.
        pyarrow_filters (Optional[PYARROW_FILTER], optional): A pyarrow expression used to filter
            specific theme type pair. Defaults to None.
        ignore_cache (bool, optional): Whether to ignore precalculated geoparquet files or not.
            Defaults to False.
        working_directory (Union[str, Path], optional): Directory where to save
            the downloaded `*.parquet` files. Defaults to "files".
        verbosity_mode (Literal["silent", "transient", "verbose"], optional): Set progress
            verbosity mode. Can be one of: silent, transient and verbose. Silent disables
            output completely. Transient tracks progress, but removes output after finished.
            Verbose leaves all progress outputs in the stdout. Defaults to "transient".
        max_workers (Optional[int], optional): Max number of multiprocessing workers used to
            process the dataset. Defaults to None.
        places_use_primary_category_only (bool, optional): Whether to use only the primary category
            for places. Defaults to False.
        places_minimal_confidence (float, optional): Minimal confidence level for the places
            dataset. Defaults to 0.75.

    Returns:
        gpd.GeoDataFrame: GeoDataFrame with Overture Maps features.
    """
    return convert_geometry_to_wide_form_geodataframe(
        theme=theme,
        type=type,
        geometry_filter=box(*bbox),
        release=release,
        include_all_possible_columns=include_all_possible_columns,
        hierarchy_depth=hierarchy_depth,
        pyarrow_filters=pyarrow_filters,
        ignore_cache=ignore_cache,
        working_directory=working_directory,
        verbosity_mode=verbosity_mode,
        max_workers=max_workers,
        places_use_primary_category_only=places_use_primary_category_only,
        places_minimal_confidence=places_minimal_confidence,
    )


@overload
def convert_bounding_box_to_wide_form_geodataframe_for_all_types(
    bbox: tuple[float, float, float, float],
    *,
    include_all_possible_columns: bool = True,
    hierarchy_depth: Optional[Union[int, list[Optional[int]]]] = None,
    pyarrow_filters: Optional[list[Optional[PYARROW_FILTER]]] = None,
    ignore_cache: bool = False,
    working_directory: Union[str, Path] = "files",
    verbosity_mode: VERBOSITY_MODE = "transient",
    max_workers: Optional[int] = None,
    places_use_primary_category_only: bool = False,
    places_minimal_confidence: float = 0.75,
) -> gpd.GeoDataFrame: ...


@overload
def convert_bounding_box_to_wide_form_geodataframe_for_all_types(
    bbox: tuple[float, float, float, float],
    release: str,
    *,
    include_all_possible_columns: bool = True,
    hierarchy_depth: Optional[Union[int, list[Optional[int]]]] = None,
    pyarrow_filters: Optional[list[Optional[PYARROW_FILTER]]] = None,
    ignore_cache: bool = False,
    working_directory: Union[str, Path] = "files",
    verbosity_mode: VERBOSITY_MODE = "transient",
    max_workers: Optional[int] = None,
    places_use_primary_category_only: bool = False,
    places_minimal_confidence: float = 0.75,
) -> gpd.GeoDataFrame: ...


@overload
def convert_bounding_box_to_wide_form_geodataframe_for_all_types(
    bbox: tuple[float, float, float, float],
    release: Optional[str] = None,
    *,
    include_all_possible_columns: bool = True,
    hierarchy_depth: Optional[Union[int, list[Optional[int]]]] = None,
    pyarrow_filters: Optional[list[Optional[PYARROW_FILTER]]] = None,
    ignore_cache: bool = False,
    working_directory: Union[str, Path] = "files",
    verbosity_mode: VERBOSITY_MODE = "transient",
    max_workers: Optional[int] = None,
    places_use_primary_category_only: bool = False,
    places_minimal_confidence: float = 0.75,
) -> gpd.GeoDataFrame: ...


def convert_bounding_box_to_wide_form_geodataframe_for_all_types(
    bbox: tuple[float, float, float, float],
    release: Optional[str] = None,
    *,
    include_all_possible_columns: bool = True,
    hierarchy_depth: Optional[Union[int, list[Optional[int]]]] = None,
    pyarrow_filters: Optional[list[Optional[PYARROW_FILTER]]] = None,
    ignore_cache: bool = False,
    working_directory: Union[str, Path] = "files",
    verbosity_mode: VERBOSITY_MODE = "transient",
    max_workers: Optional[int] = None,
    places_use_primary_category_only: bool = False,
    places_minimal_confidence: float = 0.75,
) -> gpd.GeoDataFrame:
    """
    Get GeoDataFrame for a given bounding box in a wide format for all types.

    Automatically downloads Overture Maps dataset for a given release and all available theme/types
    in a concurrent manner and returns a single GeoDataFrame as a result with multiple columns based
    on dataset schema.

    Args:
        bbox (tuple[float, float, float, float]): Bounding box used to filter data.
            Order of values: xmin, ymin, xmax, ymax.
        release (Optional[str], optional): Release version. If not provided, will automatically load
            newest available release version. Defaults to None.
        include_all_possible_columns (bool, optional): Whether to have always the same list of
            columns in the resulting file. This ensures that always the same set of columns is
            returned for a given release for different regions. This also means, that some columns
            might be all filled with a False value. Defaults to True.
        hierarchy_depth (Optional[Union[int, list[Optional[int]]]], optional): Depth used to
            calculate how many hierarchy columns should be used to generate the wide form of
            the data. Can be a single integer or a list of integers. If None, will use all
            available columns. Defaults to None.
        pyarrow_filters (Optional[list[Optional[PYARROW_FILTER]]], optional): A list of pyarrow
            expressions used to filter specific theme type pair. Must be the same length as the list
            of theme type pairs. Defaults to None.
        ignore_cache (bool, optional): Whether to ignore precalculated geoparquet files or not.
            Defaults to False.
        working_directory (Union[str, Path], optional): Directory where to save
            the downloaded `*.parquet` files. Defaults to "files".
        verbosity_mode (Literal["silent", "transient", "verbose"], optional): Set progress
            verbosity mode. Can be one of: silent, transient and verbose. Silent disables
            output completely. Transient tracks progress, but removes output after finished.
            Verbose leaves all progress outputs in the stdout. Defaults to "transient".
        max_workers (Optional[int], optional): Max number of multiprocessing workers used to
            process the dataset. Defaults to None.
        places_use_primary_category_only (bool, optional): Whether to use only the primary category
            for places. Defaults to False.
        places_minimal_confidence (float, optional): Minimal confidence level for the places
            dataset. Defaults to 0.75.

    Returns:
        gpd.GeoDataFrame: GeoDataFrame with Overture Maps features.
    """
    return convert_geometry_to_wide_form_geodataframe_for_all_types(
        geometry_filter=box(*bbox),
        release=release,
        include_all_possible_columns=include_all_possible_columns,
        hierarchy_depth=hierarchy_depth,
        pyarrow_filters=pyarrow_filters,
        ignore_cache=ignore_cache,
        working_directory=working_directory,
        verbosity_mode=verbosity_mode,
        max_workers=max_workers,
        places_use_primary_category_only=places_use_primary_category_only,
        places_minimal_confidence=places_minimal_confidence,
    )


@overload
def convert_bounding_box_to_wide_form_geodataframe_for_multiple_types(
    theme_type_pairs: list[tuple[str, str]],
    bbox: tuple[float, float, float, float],
    *,
    include_all_possible_columns: bool = True,
    hierarchy_depth: Optional[Union[int, list[Optional[int]]]] = None,
    pyarrow_filters: Optional[list[Optional[PYARROW_FILTER]]] = None,
    ignore_cache: bool = False,
    working_directory: Union[str, Path] = "files",
    verbosity_mode: VERBOSITY_MODE = "transient",
    max_workers: Optional[int] = None,
    places_use_primary_category_only: bool = False,
    places_minimal_confidence: float = 0.75,
) -> gpd.GeoDataFrame: ...


@overload
def convert_bounding_box_to_wide_form_geodataframe_for_multiple_types(
    theme_type_pairs: list[tuple[str, str]],
    bbox: tuple[float, float, float, float],
    release: str,
    *,
    include_all_possible_columns: bool = True,
    hierarchy_depth: Optional[Union[int, list[Optional[int]]]] = None,
    pyarrow_filters: Optional[list[Optional[PYARROW_FILTER]]] = None,
    ignore_cache: bool = False,
    working_directory: Union[str, Path] = "files",
    verbosity_mode: VERBOSITY_MODE = "transient",
    max_workers: Optional[int] = None,
    places_use_primary_category_only: bool = False,
    places_minimal_confidence: float = 0.75,
) -> gpd.GeoDataFrame: ...


@overload
def convert_bounding_box_to_wide_form_geodataframe_for_multiple_types(
    theme_type_pairs: list[tuple[str, str]],
    bbox: tuple[float, float, float, float],
    release: Optional[str] = None,
    *,
    include_all_possible_columns: bool = True,
    hierarchy_depth: Optional[Union[int, list[Optional[int]]]] = None,
    pyarrow_filters: Optional[list[Optional[PYARROW_FILTER]]] = None,
    ignore_cache: bool = False,
    working_directory: Union[str, Path] = "files",
    verbosity_mode: VERBOSITY_MODE = "transient",
    max_workers: Optional[int] = None,
    places_use_primary_category_only: bool = False,
    places_minimal_confidence: float = 0.75,
) -> gpd.GeoDataFrame: ...


def convert_bounding_box_to_wide_form_geodataframe_for_multiple_types(
    theme_type_pairs: list[tuple[str, str]],
    bbox: tuple[float, float, float, float],
    release: Optional[str] = None,
    *,
    include_all_possible_columns: bool = True,
    hierarchy_depth: Optional[Union[int, list[Optional[int]]]] = None,
    pyarrow_filters: Optional[list[Optional[PYARROW_FILTER]]] = None,
    ignore_cache: bool = False,
    working_directory: Union[str, Path] = "files",
    verbosity_mode: VERBOSITY_MODE = "transient",
    max_workers: Optional[int] = None,
    places_use_primary_category_only: bool = False,
    places_minimal_confidence: float = 0.75,
) -> gpd.GeoDataFrame:
    """
    Get GeoDataFrame for a given bounding box in a wide format for multiple types.

    Automatically downloads Overture Maps dataset for a given release and theme/type pairs
    in a concurrent manner and returns a single GeoDataFrame as a result with multiple columns based
    on dataset schema.

    Args:
        theme_type_pairs (list[tuple[str, str]]): Pairs of themes and types of the dataset.
        bbox (tuple[float, float, float, float]): Bounding box used to filter data.
            Order of values: xmin, ymin, xmax, ymax.
        release (Optional[str], optional): Release version. If not provided, will automatically load
            newest available release version. Defaults to None.
        include_all_possible_columns (bool, optional): Whether to have always the same list of
            columns in the resulting file. This ensures that always the same set of columns is
            returned for a given release for different regions. This also means, that some columns
            might be all filled with a False value. Defaults to True.
        hierarchy_depth (Optional[Union[int, list[Optional[int]]]], optional): Depth used to
            calculate how many hierarchy columns should be used to generate the wide form of
            the data. Can be a single integer or a list of integers. If None, will use all
            available columns. Defaults to None.
        pyarrow_filters (Optional[list[Optional[PYARROW_FILTER]]], optional): A list of pyarrow
            expressions used to filter specific theme type pair. Must be the same length as the list
            of theme type pairs. Defaults to None.
        ignore_cache (bool, optional): Whether to ignore precalculated geoparquet files or not.
            Defaults to False.
        working_directory (Union[str, Path], optional): Directory where to save
            the downloaded `*.parquet` files. Defaults to "files".
        verbosity_mode (Literal["silent", "transient", "verbose"], optional): Set progress
            verbosity mode. Can be one of: silent, transient and verbose. Silent disables
            output completely. Transient tracks progress, but removes output after finished.
            Verbose leaves all progress outputs in the stdout. Defaults to "transient".
        max_workers (Optional[int], optional): Max number of multiprocessing workers used to
            process the dataset. Defaults to None.
        places_use_primary_category_only (bool, optional): Whether to use only the primary category
            for places. Defaults to False.
        places_minimal_confidence (float, optional): Minimal confidence level for the places
            dataset. Defaults to 0.75.

    Returns:
        gpd.GeoDataFrame: GeoDataFrame with Overture Maps features.
    """
    return convert_geometry_to_wide_form_geodataframe_for_multiple_types(
        theme_type_pairs=theme_type_pairs,
        geometry_filter=box(*bbox),
        release=release,
        include_all_possible_columns=include_all_possible_columns,
        hierarchy_depth=hierarchy_depth,
        pyarrow_filters=pyarrow_filters,
        ignore_cache=ignore_cache,
        working_directory=working_directory,
        verbosity_mode=verbosity_mode,
        max_workers=max_workers,
        places_use_primary_category_only=places_use_primary_category_only,
        places_minimal_confidence=places_minimal_confidence,
    )
